from typing import TYPE_CHECKING, Any, AsyncGenerator

from authlib.integrations.base_client.errors import OAuthError
from authlib.jose.errors import BadSignatureError, DecodeError, ExpiredTokenError
from fastapi import Depends, HTTPException, Path, status
from fastapi.requests import Request
from fastapi.security import HTTPBearer
from fastapi.security.http import HTTPAuthorizationCredentials
from rgwadmin import RGWAdmin
from sqlalchemy.ext.asyncio import AsyncSession

from app.ceph.rgw import rgw, s3_resource
from app.core.security import decode_token, oauth
from app.crud.crud_bucket import CRUDBucket
from app.crud.crud_bucket_permission import CRUDBucketPermission
from app.crud.crud_user import CRUDUser
from app.db.session import SessionAsync as Session
from app.models.bucket import Bucket
from app.models.user import User
from app.schemas.security import JWTToken

if TYPE_CHECKING:
    from boto3.resources.base import ServiceResource
else:
    ServiceResource = object

bearer_token = HTTPBearer(description="JWT Token")


class LoginException(Exception):
    def __init__(self, error_source: str):
        self.error_source = error_source


def get_rgw_admin() -> RGWAdmin:
    return rgw  # pragma: no cover


def get_s3_resource() -> ServiceResource:
    return s3_resource  # pragma: no cover


async def get_db() -> AsyncGenerator:
    """
    Get a Session with the database.

    FastAPI Dependency Injection Function.

    Returns
    -------
    db : AsyncGenerator
        Async session object with the database
    """
    async with Session() as db:
        yield db


def decode_bearer_token(
    token: HTTPAuthorizationCredentials = Depends(bearer_token),
) -> JWTToken:
    """
    Get the decoded JWT or reject request if it is not valid.

    FastAPI Dependency Injection Function.

    Parameters
    ----------
    token : fastapi.security.http.HTTPAuthorizationCredentials
        Bearer token sent with the HTTP request. Dependency Injection.

    Returns
    -------
    token : app.schemas.security.JWTToken
        The verified and decoded JWT.
    """
    try:
        return JWTToken(**decode_token(token.credentials))
    except ExpiredTokenError:  # pragma: no cover
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="JWT Signature has expired")
    except (DecodeError, BadSignatureError):
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Malformed JWT")


async def get_current_user(token: JWTToken = Depends(decode_bearer_token), db: AsyncSession = Depends(get_db)) -> User:
    """
    Get the current user from the database based on the JWT.

    FastAPI Dependency Injection Function.

    Parameters
    ----------
    token : app.schemas.security.JWTToken
        The verified and decoded JWT.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.

    Returns
    -------
    user : app.models.user.User
        User associated with the JWT sent with the HTTP request.
    """
    user = await CRUDUser.get(db, token.sub)
    if user:
        return user
    raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="User not found")


async def get_user_by_path_uid(
    uid: str = Path(
        default=..., description="UID of a user", example="28c5353b8bb34984a8bd4169ba94c606", max_length=64
    ),
    db: AsyncSession = Depends(get_db),
    token: JWTToken = Depends(decode_bearer_token),
) -> User:
    """
    Get the user from the database with the given uid.
    Reject the request if the current user is not the same as the requested one.

    FastAPI Dependency Injection Function.

    Parameters
    ----------
    uid : str
        The uid of a user. URL path parameter.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    token : app.schemas.security.JWTToken
        Decoded JWT sent with the HTTP request.

    Returns
    -------
    user : app.models.user.User
        User with the given uid.

    """
    user = await CRUDUser.get(db, uid)
    if user:
        if user.uid == token.sub:
            return user
        else:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="A user can only access himself",
            )
    raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="User not found")


async def get_current_bucket(
    bucket_name: str = Path(..., description="Name of bucket", example="test-bucket", max_length=63, min_length=3),
    db: AsyncSession = Depends(get_db),
    current_user: User = Depends(get_current_user),
) -> Bucket:
    """
    Get the Bucket from the database based on the name in the path.
    Reject the request if user has no READ permission for this bucket.

    FastAPI Dependency Injection Function

    Parameters
    ----------
    bucket_name : str
        Name of a bucket. URL Path Parameter.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : app.models.user.User
        User associated with the JWT sent with the HTTP request. Dependency Injection

    Returns
    -------
    bucket : app.models.bucket.Bucket
        Bucket with the given name.
    """
    bucket = await CRUDBucket.get(db, bucket_name)
    if bucket is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="Bucket not found")
    elif not bucket.public and not await CRUDBucketPermission.check_permission(db, bucket_name, current_user.uid):
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail="No rights for this bucket")
    return bucket


async def get_authorized_user_for_permission(
    bucket: Bucket = Depends(get_current_bucket),
    uid: str = Path(
        default=..., description="UID of a user", example="28c5353b8bb34984a8bd4169ba94c606", max_length=64
    ),
    db: AsyncSession = Depends(get_db),
    current_user: User = Depends(get_current_user),
) -> User:
    """
    Get the user for viewing and deleting bucket permissions.\n
    Only the owner of a bucket and grantee can do this.

    Parameters
    ----------
    bucket : app.models.bucket.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    uid : str
        The uid of a user. URL path parameter.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : app.models.user.User
        Current user. Dependency Injection.

    Returns
    -------
    user : app.models.user.User
        Authorized user for bucket permission. Dependency Injection.
    """
    user = await CRUDUser.get(db, uid)
    if user is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="User not found")
    elif current_user != user and current_user.uid != bucket.owner_id:
        raise HTTPException(
            status.HTTP_403_FORBIDDEN, detail="Only the owner or the grantee can access a bucket permission"
        )
    return user


async def get_userinfo_from_access_token(request: Request) -> dict[str, Any]:  # pragma: no cover
    """
    Get the userinfo from the OAuth2 userinfo endpoint with the access token.

    Parameters
    ----------
    request : starlette.requests.Request
        Raw Request object.

    Returns
    -------
    userinfo : dict[str, Any]
        Info about the corresponding user.
    """
    try:
        if "error" in request.query_params.keys():
            # if there is an error in the login flow, like a canceld login request, then notify the client
            raise LoginException(error_source=request.query_params["error"])
        claims = await oauth.lifescience.authorize_access_token(request)
        # ID token doesn't have all necessary information, call userinfo endpoint
        return await oauth.lifescience.userinfo(token=claims)
    except OAuthError:
        # if there is an error in the oauth flow, like an expired token, then notify the client
        raise LoginException(error_source="oidc")
