import json
from typing import TYPE_CHECKING

from fastapi import APIRouter, Body, Depends, HTTPException, status
from sqlalchemy.ext.asyncio import AsyncSession

from app.api.dependencies import (
    get_authorized_user_for_permission,
    get_current_bucket,
    get_current_user,
    get_db,
    get_s3_resource,
    get_user_by_path_uid,
)
from app.crud.crud_bucket_permission import CRUDBucketPermission, DuplicateError
from app.crud.crud_user import CRUDUser
from app.models.bucket import Bucket as BucketDB
from app.models.user import User as UserDB
from app.schemas.bucket_permission import BucketPermissionIn as PermissionSchemaIn
from app.schemas.bucket_permission import BucketPermissionOut as PermissionSchemaOut
from app.schemas.bucket_permission import BucketPermissionParameters as PermissionParametersSchema

router = APIRouter(prefix="/permissions", tags=["BucketPermissions"])

if TYPE_CHECKING:
    from mypy_boto3_s3.service_resource import S3ServiceResource
else:
    S3ServiceResource = object


@router.get(
    "/bucket/{bucket_name}/user/{uid}",
    response_model=PermissionSchemaOut,
    summary="Get permission for bucket and user combination.",
    response_model_exclude_none=True,
)
async def get_permission_for_bucket(
    bucket: BucketDB = Depends(get_current_bucket),
    db: AsyncSession = Depends(get_db),
    user: UserDB = Depends(get_authorized_user_for_permission),
) -> PermissionSchemaOut:
    """
    Get the bucket permissions for the specific combination of bucket and user.\n
    The owner of the bucket and the grantee of the permission can view it.
    \f
    Parameters
    ----------
    bucket : app.models.bucket.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    user : app.models.user.User
        User with the uid in the URL. Dependency Injection.

    Returns
    -------
    permissions : app.schemas.bucket_permission.BucketPermissionOut
        Permission for this bucket and user combination.
    """
    bucket_permission = await CRUDBucketPermission.get(db, bucket.name, user.uid)
    if bucket_permission:
        return PermissionSchemaOut.from_db_model(
            bucket_permission, uid=user.uid, grantee_display_name=user.display_name
        )
    raise HTTPException(
        status.HTTP_404_NOT_FOUND,
        detail=f"Permission for combination of bucket={bucket.name} and user={user.uid} doesn't exists",
    )


@router.delete(
    "/bucket/{bucket_name}/user/{uid}",
    status_code=status.HTTP_204_NO_CONTENT,
    summary="Delete a bucket permission",
)
async def delete_permission_for_bucket(
    bucket: BucketDB = Depends(get_current_bucket),
    db: AsyncSession = Depends(get_db),
    user: UserDB = Depends(get_authorized_user_for_permission),
    s3: S3ServiceResource = Depends(get_s3_resource),
) -> None:
    """
    Delete the bucket permissions for the specific combination of bucket and user.\n
    The owner of the bucket and the grantee of the permission can delete it.
    \f
    Parameters
    ----------
    bucket : app.models.bucket.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    user : app.models.user.User
        User with the uid in the URL. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.

    Returns
    -------
    permissions : app.schemas.bucket_permission.BucketPermissionOut
        Permission for this bucket and user combination.
    """
    bucket_permission = await CRUDBucketPermission.get(db, bucket.name, user.uid)
    if bucket_permission is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail=f"Permission for combination of bucket={bucket.name} and user={user.uid} doesn't exists",
        )
    await CRUDBucketPermission.delete(db, bucket_permission)
    bucket_permission_schema = PermissionSchemaOut.from_db_model(bucket_permission, user.uid, user.display_name)
    s3_policy = s3.Bucket(bucket_permission_schema.bucket_name).Policy()
    policy = json.loads(s3_policy.policy)
    policy["Statement"] = [
        stmt for stmt in policy["Statement"] if stmt["Sid"] != bucket_permission_schema.to_hash(user.uid)
    ]
    s3_policy.put(Policy=json.dumps(policy))


@router.get(
    "/bucket/{bucket_name}",
    response_model=list[PermissionSchemaOut],
    summary="Get all permissions for a bucket.",
    response_model_exclude_none=True,
)
async def list_permissions_per_bucket(
    bucket: BucketDB = Depends(get_current_bucket),
    db: AsyncSession = Depends(get_db),
    current_user: UserDB = Depends(get_current_user),
) -> list[PermissionSchemaOut]:
    """
    List all the bucket permissions for the given bucket.
    \f
    Parameters
    ----------
    bucket : app.models.bucket.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : app.models.user.User
        Current user. Dependency Injection.

    Returns
    -------
    permissions : list[app.schemas.bucket_permission.BucketPermissionOut]
        List of all permissions for this bucket.
    """
    if not await CRUDBucketPermission.check_permission(db, bucket.name, current_user.uid, only_own=True):
        raise HTTPException(status.HTTP_403_FORBIDDEN, "You can only view your own bucket permissions")
    bucket_permissions = await CRUDBucketPermission.get_permissions_for_bucket(db, bucket.name)
    return [PermissionSchemaOut.from_db_model(p) for p in bucket_permissions]


@router.get(
    "/user/{uid}",
    response_model=list[PermissionSchemaOut],
    summary="Get all permissions for a user.",
    response_model_exclude_none=True,
)
async def list_permissions_per_user(
    user: UserDB = Depends(get_user_by_path_uid), db: AsyncSession = Depends(get_db)
) -> list[PermissionSchemaOut]:
    """
    List all the bucket permissions for the given user.
    \f
    Parameters
    ----------
    user :  : app.models.user.User
        User with given uid. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.

    Returns
    -------
    permissions : list[app.schemas.bucket_permission.BucketPermissionOut]
        List of all permissions for this user.
    """
    bucket_permissions = await CRUDBucketPermission.get_permissions_for_user(db, user.uid)
    return [
        PermissionSchemaOut.from_db_model(p, uid=user.uid, grantee_display_name=user.display_name)
        for p in bucket_permissions
    ]


@router.post(
    "/",
    response_model=PermissionSchemaOut,
    status_code=status.HTTP_201_CREATED,
    summary="Create a permission.",
    response_model_exclude_none=True,
)
async def create_permission(
    permission: PermissionSchemaIn = Body(..., description="Permission to create"),
    db: AsyncSession = Depends(get_db),
    current_user: UserDB = Depends(get_current_user),
    s3: S3ServiceResource = Depends(get_s3_resource),
) -> PermissionSchemaOut:
    """
    Create a permission for a bucket and user.
    \f
    Parameters
    ----------
    permission : app.schemas.bucket_permission.BucketPermissionOut
        Information about the permission which should be created. HTTP Body parameter.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : app.models.user.User
        Current user. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.

    Returns
    -------
    permissions : app.schemas.bucket_permission.BucketPermissionOut
        Newly created permission.
    """
    await get_current_bucket(permission.bucket_name, db=db, current_user=current_user)
    try:
        permission_db = await CRUDBucketPermission.create(db, permission)
        grantee = await CRUDUser.get(db, permission.uid)
        if grantee is None:  # pragma: no cover
            raise KeyError()
    except ValueError:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST, detail="The owner of the bucket can't get any more permissions"
        )
    except DuplicateError:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            detail=f"Permission for combination of bucket={permission.bucket_name} and user={permission.uid} already exists",  # noqa:E501
        )
    except KeyError:
        raise HTTPException(status.HTTP_404_NOT_FOUND, detail=f"User with uid={permission.uid} not found")
    s3_policy = s3.Bucket(permission.bucket_name).Policy()
    old_policy = s3_policy.policy
    json_policy = json.loads(old_policy)
    json_policy["Statement"] += permission.map_to_bucket_policy_statement(permission_db.user_id)
    new_policy = json.dumps(json_policy)
    s3_policy.put(Policy=new_policy)

    return PermissionSchemaOut.from_db_model(permission_db, uid=grantee.uid, grantee_display_name=grantee.display_name)


@router.put(
    "/bucket/{bucket_name}/user/{uid}",
    status_code=status.HTTP_200_OK,
    response_model=PermissionSchemaOut,
    summary="Update a bucket permission",
    response_model_exclude_none=True,
)
async def update_permission(
    permission_parameters: PermissionParametersSchema = Body(..., description="Permission to create"),
    user: UserDB = Depends(get_authorized_user_for_permission),
    bucket: BucketDB = Depends(get_current_bucket),
    db: AsyncSession = Depends(get_db),
    current_user: UserDB = Depends(get_current_user),
    s3: S3ServiceResource = Depends(get_s3_resource),
) -> PermissionSchemaOut:
    """
    Update a permission for a bucket and user.
    \f
    Parameters
    ----------
    permission_parameters : app.schemas.bucket_permission.BucketPermissionOut
        Information about the permission which should be updated. HTTP Body parameter.
    user : app.models.user.User
        User with the uid in the URL. Dependency Injection.
    bucket : app.models.bucket.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    current_user : app.models.user.User
        Current user. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.

    Returns
    -------
    permissions : app.schemas.bucket_permission.BucketPermissionOut
        Updated permission.
    """
    if not await CRUDBucketPermission.check_permission(db, bucket.name, current_user.uid, only_own=True):
        raise HTTPException(status.HTTP_403_FORBIDDEN, "You can only modify permissions on your own bucket")
    bucket_permission = await CRUDBucketPermission.get(db, bucket.name, user.uid)

    if bucket_permission is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail=f"Permission for combination of bucket={bucket.name} and user={user.uid} doesn't exists",
        )
    updated_permission = await CRUDBucketPermission.update_permission(db, bucket_permission, permission_parameters)
    updated_permission_schema = PermissionSchemaOut.from_db_model(updated_permission)
    s3_policy = s3.Bucket(updated_permission_schema.bucket_name).Policy()
    policy = json.loads(s3_policy.policy)
    policy["Statement"] = [
        stmt for stmt in policy["Statement"] if stmt["Sid"] != updated_permission_schema.to_hash(user.uid)
    ]
    policy["Statement"] += updated_permission_schema.map_to_bucket_policy_statement(updated_permission.user_id)
    s3_policy.put(Policy=json.dumps(policy))
    return updated_permission_schema
