import json
from functools import reduce
from typing import TYPE_CHECKING

from botocore.exceptions import ClientError
from fastapi import APIRouter, Depends, HTTPException, Path, Query, status
from sqlalchemy.ext.asyncio import AsyncSession

from app.api.dependencies import get_current_bucket, get_current_user, get_db, get_s3_resource
from app.core.config import settings
from app.crud.crud_bucket import CRUDBucket
from app.crud.crud_bucket_permission import CRUDBucketPermission
from app.models.bucket import Bucket as BucketDB
from app.models.user import User
from app.schemas.bucket import BucketIn as BucketInSchema
from app.schemas.bucket import BucketOut as BucketOutSchema
from app.schemas.bucket import S3ObjectMetaInformation

if TYPE_CHECKING:
    from mypy_boto3_s3.service_resource import ObjectSummary, S3ServiceResource
else:
    S3ServiceResource = object
    ObjectSummary = object

router = APIRouter(prefix="/buckets", tags=["Bucket"])


@router.get("/", response_model=list[BucketOutSchema], summary="List buckets of user")
async def list_buckets(
    user: User = Depends(get_current_user),
    db: AsyncSession = Depends(get_db),
    s3: S3ServiceResource = Depends(get_s3_resource),
) -> list[BucketOutSchema]:
    """
    List the buckets of the current user where the user has READ permissions for.
    \f
    Parameters
    ----------
    user : app.models.user.User
        User for which to retrieve the buckets. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.

    Returns
    -------
    buckets : list[app.schemas.bucket.BucketOut]
        All the buckets for which the user has READ permissions.
    """
    buckets = [
        BucketOutSchema(
            **{
                "description": bucket.description,
                "name": bucket.name,
                "created_at": s3.Bucket(name=bucket.name).creation_date,
                "owner": bucket.owner.uid,
                "num_objects": sum(
                    1 for obj in s3.Bucket(name=bucket.name).objects.all() if not obj.key.endswith(".s3keep")
                ),
                "size": reduce(lambda x, y: x + y.size, list(s3.Bucket(name=bucket.name).objects.all()), 0),
            }
        )
        for bucket in await CRUDBucket.get_for_user(db, user.uid)
    ]

    return buckets


@router.post(
    "/",
    response_model=BucketOutSchema,
    status_code=status.HTTP_201_CREATED,
    summary="Create a bucket for the current user",
)
async def create_bucket(
    bucket: BucketInSchema,
    user: User = Depends(get_current_user),
    db: AsyncSession = Depends(get_db),
    s3: S3ServiceResource = Depends(get_s3_resource),
) -> BucketOutSchema:
    """
    Create a bucket for the current user.\n
    The name of the bucket has some constraints.
    For more information see the [Ceph documentation](https://docs.ceph.com/en/quincy/radosgw/s3/bucketops/#constraints)
    \f
    Parameters
    ----------
    bucket : app.schemas.bucket.BucketIn
        Information about the bucket to create. HTTP Body
    user : app.models.user.User
        Current user who will be the owner of the newly created bucket. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.

    Returns
    -------
    bucket : app.schemas.bucket.BucketOut
        The newly created bucket.
    """

    db_bucket = await CRUDBucket.create(db, bucket, user.uid)
    if db_bucket is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Bucket name is already taken",
        )
    s3_bucket = s3.Bucket(db_bucket.name)
    s3_bucket.create()
    # Add basic permission to the user for getting, creating and deleting objects in the bucket.
    bucket_policy = json.dumps(
        {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "ProxyOwnerPerm",
                    "Effect": "Allow",
                    "Principal": {"AWS": [f"arn:aws:iam:::user/{settings.CEPH_USERNAME}"]},
                    "Action": ["s3:GetObject"],
                    "Resource": [f"arn:aws:s3:::{db_bucket.name}/*"],
                },
                {
                    "Sid": "PseudoOwnerPerm",
                    "Effect": "Allow",
                    "Principal": {"AWS": [f"arn:aws:iam:::user/{user.uid}"]},
                    "Action": ["s3:GetObject", "s3:DeleteObject", "s3:PutObject", "s3:ListBucket"],
                    "Resource": [f"arn:aws:s3:::{db_bucket.name}/*", f"arn:aws:s3:::{db_bucket.name}"],
                },
            ],
        }
    )
    s3_bucket.Policy().put(Policy=bucket_policy)
    return BucketOutSchema(
        **{
            "description": db_bucket.description,
            "name": db_bucket.name,
            "created_at": s3.Bucket(name=db_bucket.name).creation_date,
            "owner": db_bucket.owner.uid,
            "num_objects": 0,
            "size": 0,
        }
    )


@router.get("/{bucket_name}", response_model=BucketOutSchema, summary="Get a bucket by its name")
async def get_bucket(
    bucket: BucketDB = Depends(get_current_bucket), s3: S3ServiceResource = Depends(get_s3_resource)
) -> BucketOutSchema:
    """
    Get a bucket by its name if the current user has READ permissions for the bucket.
    \f
    Parameters
    ----------
    bucket : app.models.bucket.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.

    Returns
    -------
    bucket : app.schemas.bucket.BucketOut
        Bucket with the provided name.
    """
    s3bucket = s3.Bucket(name=bucket.name)
    objects: list[ObjectSummary] = list(s3bucket.objects.all())
    return BucketOutSchema(
        **{
            "description": bucket.description,
            "name": bucket.name,
            "created_at": s3bucket.creation_date,
            "owner": bucket.owner.uid,
            "num_objects": sum(1 for obj in objects if not obj.key.endswith(".s3keep")),
            "size": reduce(lambda x, y: x + y.size, objects, 0),
        }
    )


@router.delete("/{bucket_name}", status_code=status.HTTP_204_NO_CONTENT, summary="Delete a bucket")
async def delete_bucket(
    force_delete: bool = Query(False, description="Delete even non-empty bucket"),
    bucket: BucketDB = Depends(get_current_bucket),
    user: User = Depends(get_current_user),
    db: AsyncSession = Depends(get_db),
    s3: S3ServiceResource = Depends(get_s3_resource),
) -> None:
    """
    Delete a bucket by its name. Only the owner of the bucket can delete the bucket.
    \f
    Parameters
    ----------
    force_delete : bool, default False
        Flag for deleting a non-empty bucket. Query parameter.
    bucket : app.models.bucket.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    user : app.models.user.User
        Current user. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    """
    if not await CRUDBucketPermission.check_permission(db, bucket.name, user.uid, only_own=True):
        raise HTTPException(status.HTTP_403_FORBIDDEN, "You can only delete your own buckets")
    if force_delete:
        objs = [{"Key": obj.key} for obj in s3.Bucket(bucket.name).objects.all()]
        if len(objs) > 0:
            s3.Bucket(bucket.name).delete_objects(Delete={"Objects": objs})  # type: ignore
    try:
        s3.Bucket(name=bucket.name).delete()
        await CRUDBucket.delete(db, bucket)
    except ClientError:
        raise HTTPException(status.HTTP_400_BAD_REQUEST, detail="Bucket not empty")


@router.get(
    "/{bucket_name}/objects",
    response_model=list[S3ObjectMetaInformation],
    tags=["Object"],
    summary="Get the metadata of the objects in the bucket",
)
async def get_bucket_objects(
    bucket: BucketDB = Depends(get_current_bucket),
    s3: S3ServiceResource = Depends(get_s3_resource),
    current_user: User = Depends(get_current_user),
    db: AsyncSession = Depends(get_db),
) -> list[S3ObjectMetaInformation]:
    """
    Get the metadata of the objects in the bucket.
    \f
    Parameters
    ----------
    bucket : app.models.bucket.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    current_user : app.models.user.User
        Current user. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.

    Returns
    -------
    objs : list[app.schemas.bucket.S3ObjectMetaInformation]
        Meta information about all objects in the bucket.
    """
    permission = await CRUDBucketPermission.get(db, bucket.name, current_user.uid)
    if permission is not None and permission.file_prefix is not None:
        return [
            S3ObjectMetaInformation.from_native_s3_object(obj)
            for obj in s3.Bucket(bucket.name).objects.filter(Prefix=permission.file_prefix).all()
        ]
    return [S3ObjectMetaInformation.from_native_s3_object(obj) for obj in s3.Bucket(bucket.name).objects.all()]


@router.get(
    "/{bucket_name}/objects/{object_path:path}",
    response_model=S3ObjectMetaInformation,
    tags=["Object"],
    summary="Get the metadata about a specific object",
)
async def get_bucket_object(
    bucket: BucketDB = Depends(get_current_bucket),
    s3: S3ServiceResource = Depends(get_s3_resource),
    object_path: str = Path(
        ...,
        decsription="Name of the object",
        examples={
            "normal": {
                "summary": "Normal file",
                "description": "A normal file in a bucket",
                "value": "test.txt",
            },
            "pseudo-folder": {
                "summary": "Pseudo-folder file",
                "description": "A file in a pseudo folder",
                "value": "pseudo/sub/folder/test.txt",
            },
        },
    ),
    current_user: User = Depends(get_current_user),
    db: AsyncSession = Depends(get_db),
) -> S3ObjectMetaInformation:
    """
    Get the metadata of a specific object in a bucket.
    \f
    Parameters
    ----------
    bucket : app.models.bucket.Bucket
        Bucket with the name provided in the URL path. Dependency Injection.
    s3 : boto3_type_annotations.s3.ServiceResource
        S3 Service to perform operations on buckets in Ceph. Dependency Injection.
    object_path : str
        Key of a specific object in the bucket. URL Path Parameter.
    current_user : app.models.user.User
        Current user. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.

    Returns
    -------
    objs : app.schemas.bucket.S3ObjectMetaInformation
        Meta information about the specific object in the bucket.
    """
    permission = await CRUDBucketPermission.get(db, bucket.name, current_user.uid)
    try:
        if permission is None or permission.file_prefix is None or object_path.startswith(permission.file_prefix):
            obj = s3.ObjectSummary(bucket_name=bucket.name, key=object_path)
            return S3ObjectMetaInformation.from_native_s3_object(obj)
        raise HTTPException(status.HTTP_403_FORBIDDEN, detail="No rights for this object.")
    except ClientError:
        raise HTTPException(status.HTTP_404_NOT_FOUND, detail="Object not found")
