from typing import Any

from fastapi import APIRouter, Depends, Request, Response, status
from fastapi.responses import RedirectResponse
from rgwadmin import RGWAdmin
from sqlalchemy.ext.asyncio import AsyncSession

from app.api.dependencies import LoginException, get_db, get_rgw_admin, get_userinfo_from_access_token
from app.core.config import settings
from app.core.security import create_access_token, oauth
from app.crud.crud_user import CRUDUser
from app.models.user import User

router = APIRouter(prefix="/auth", tags=["Auth"])


@router.get("/login", include_in_schema=False)
async def login(request: Request) -> RedirectResponse:
    """
    Redirect route to OIDC provider.
    \f
    Parameters
    ----------
    request : fastapi.requests.Request
        Raw request object.

    Returns
    -------
    response : fastapi.responses.RedirectResponse
        Redirect response to right OAuth2 endpoint
    """
    # Clear session to prevent an overflow
    request.session.clear()
    # construct absolute url for callback
    base_url = str(request.base_url)[:-1]
    if settings.SSL_TERMINATION:  # pragma: no cover
        base_url = "https" + base_url[4:]
    redirect_uri = base_url + router.prefix + "/callback"
    return await oauth.lifescience.authorize_redirect(request, redirect_uri)


@router.get(
    "/callback",
    response_class=RedirectResponse,
    status_code=status.HTTP_302_FOUND,
    summary="Life Science Login Callback",
    responses={
        status.HTTP_302_FOUND: {
            "headers": {
                "Set-Cookie": {
                    "description": "JWT for accessing the API",
                    "schema": {
                        "type": "string",
                        "example": "bearer=fake-jwt-cookie; Domain=localhost; expired=Wed, 05 Jan 2022 "
                        "09:00:00 GMT; Path=/; SameSite=strict; Secure",
                    },
                }
            }
        }
    },
)
async def login_callback(
    response: RedirectResponse,
    user_info: dict[str, Any] = Depends(get_userinfo_from_access_token),
    db: AsyncSession = Depends(get_db),
    rgw: RGWAdmin = Depends(get_rgw_admin),
) -> str:
    """
    Callback for the Life Science Identity Provider.\n
    To start the login process visit the route [login route](/api/auth/login/)

    If the user is already known to the system, then a JWT token will be created and sent via the 'set-cookie' header.
    The key for this Cookie is 'bearer'.\n
    If the user is new, he will be created and then a JWT token is issued.\n
    This JWT has to be sent to all authorized endpoints via the HTTPBearer scheme.
    \f
    Parameters
    ----------
    response : fastapi.responses.RedirectResponse
        Response which will hold the JWT cookie.
    user_info : dict[str, Any]
        Get the userinfo with OAuth2. Dependency Injection.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.
    rgw : rgwadmin.RGWAdmin
        RGW admin interface to manage Ceph's object store. Dependency Injection.

    Returns
    -------
    path : str
        Redirect path after successful login.
    """
    try:
        lifescience_id = (
            user_info["voperson_id"] if isinstance(user_info["voperson_id"], str) else user_info["voperson_id"][0]
        )
        uid = lifescience_id.split("@")[0]
        user = await CRUDUser.get(db, uid)
        if user is None:
            new_user = User(uid=uid, display_name=user_info["name"])
            user = await CRUDUser.create(db, new_user)
            rgw.create_user(uid=user.uid, max_buckets=-1, display_name=user.display_name)
        token = create_access_token(uid)
        response.set_cookie(
            key="bearer",
            value=token,
            samesite="strict",
            max_age=settings.JWT_TOKEN_EXPIRE_MINUTES,
            secure=True,
            domain=settings.DOMAIN,
        )
    except Exception:  # pragma: no cover
        raise LoginException(error_source="server")
    return "/"


def login_exception_handler(request: Request, exc: LoginException) -> Response:
    """
    Exception handler for all kinds of login errors.

    Parameters
    ----------
    request : fastapi.Request
        Original request where the exception occurred.
    exc : LoginException
        The exception that was raised.

    Returns
    -------
    redirect : fastapi.Response
        Redirect to base URL with error as query parameter
    """
    return RedirectResponse(f"/?login_error={exc.error_source}", status_code=status.HTTP_302_FOUND)
