from fastapi import APIRouter, Depends, HTTPException, Path, Query, status
from rgwadmin import RGWAdmin
from rgwadmin.exceptions import RGWAdminException
from sqlalchemy.ext.asyncio import AsyncSession

from app.api.dependencies import get_current_user, get_db, get_rgw_admin, get_user_by_path_uid
from app.crud.crud_user import CRUDUser
from app.models.user import User as UserDB
from app.schemas.user import S3Key
from app.schemas.user import User as UserSchema

router = APIRouter(prefix="/users", tags=["User"])


@router.get("/me", response_model=UserSchema, summary="Get the logged in user")
def get_logged_in_user(
    current_user: UserDB = Depends(get_current_user),
) -> UserDB:
    """
    Return the user associated with the used JWT.
    \f
    Parameters
    ----------
    current_user : app.models.user.User
        User from the database associated to the used JWT. Dependency injection.

    Returns
    -------
    current_user : app.models.user.User
        User associated to used JWT.
    """
    return current_user


@router.get("/", response_model=list[UserSchema], summary="Search for users by their name")
async def search_users(
    name_like: str = Query(..., min_length=3, max_length=30),
    db: AsyncSession = Depends(get_db),
) -> list[UserDB]:
    """
    Return the users that have a specific substring in their name.
    \f
    Parameters
    ----------
    name_like : string
        Substring of a name to search users for. Query Parameter.
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on. Dependency Injection.

    Returns
    -------
    users: list[app.models.user.User]
        Users who have the substring in their name.
    """
    return await CRUDUser.search_for_name(db, name_like)


@router.get("/{uid}", response_model=UserSchema, summary="Get a user by its uid")
def get_user(user: UserDB = Depends(get_user_by_path_uid)) -> UserDB:
    """
    Return the user with the specific uid. A user can only view himself.
    \f
    Parameters
    ----------
    user : app.models.user.User
        User with given uid. Dependency Injection.
    Returns
    -------
    user : app.models.user.User
        User with given uid.
    """
    return user


@router.get(
    "/{uid}/keys",
    response_model=list[S3Key],
    tags=["Key"],
    summary="Get the S3 Access keys from a user",
)
def get_user_keys(rgw: RGWAdmin = Depends(get_rgw_admin), user: UserDB = Depends(get_user_by_path_uid)) -> list[S3Key]:
    """
    Get all the S3 Access keys for a specific user.
    \f
    Parameters
    ----------
    rgw : rgwadmin.RGWAdmin
        RGW admin interface to manage Ceph's object store. Dependency Injection.
    user : app.models.user.User
        User with given uid. Dependency Injection.

    Returns
    -------
    keys : list(app.schemas.user.S3Key)
        All S3 keys from the user.
    """
    return [S3Key(**key) for key in rgw.get_user(uid=user.uid, stats=False)["keys"]]


@router.post(
    "/{uid}/keys",
    response_model=S3Key,
    tags=["Key"],
    summary="Create a Access key for a user",
    status_code=status.HTTP_201_CREATED,
)
def create_user_key(rgw: RGWAdmin = Depends(get_rgw_admin), user: UserDB = Depends(get_user_by_path_uid)) -> S3Key:
    """
    Create a S3 Access key for a specific user.
    \f
    Parameters
    ----------
    rgw : rgwadmin.RGWAdmin
        RGW admin interface to manage Ceph's object store. Dependency Injection.
    user : app.models.user.User
        User with given uid. Dependency Injection.

    Returns
    -------
    key : app.schemas.user.S3Key
        Newly created S3 key.
    """
    before_keys_set = set(
        map(
            lambda key: key["access_key"],
            rgw.get_user(uid=user.uid, stats=False)["keys"],
        )
    )
    # create keys returns all keys for a user including the new one
    after_keys = rgw.create_key(uid=user.uid, key_type="s3", generate_key=True)
    new_key_id = list(set(map(lambda key: key["access_key"], after_keys)) - before_keys_set)[0]  # find ID of the key
    index = [key["access_key"] for key in after_keys].index(new_key_id)  # find new key by ID
    return S3Key(**after_keys[index])


@router.get(
    "/{uid}/keys/{access_id}",
    response_model=S3Key,
    tags=["Key"],
    summary="Get a specific S3 Access key from a user",
)
def get_user_key(
    access_id: str = Path(
        ...,
        description="ID of the S3 access key",
        example="CRJ6B037V2ZT4U3W17VC",
    ),
    rgw: RGWAdmin = Depends(get_rgw_admin),
    user: UserDB = Depends(get_user_by_path_uid),
) -> S3Key:
    """
    Get a specific S3 Access Key for a specific user.
    \f
    Parameters
    ----------
    access_id : str
        ID of the requested S3 key. URL Path Parameter.
    rgw : rgwadmin.RGWAdmin
        RGW admin interface to manage Ceph's object store. Dependency Injection.
    user : app.models.user.User
        User with given uid. Dependency Injection.

    Returns
    -------
    key : app.schemas.user.S3Key
        Requested S3 key.
    """
    keys = rgw.get_user(uid=user.uid, stats=False)["keys"]
    try:
        index = [key["access_key"] for key in keys].index(access_id)
        return S3Key(**keys[index])
    except ValueError:
        raise HTTPException(status.HTTP_404_NOT_FOUND, detail="Key not found")


@router.delete(
    "/{uid}/keys/{access_id}",
    tags=["Key"],
    summary="Delete a specific S3 Access key from a user",
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_user_key(
    access_id: str = Path(
        ...,
        description="ID of the S3 access key",
        example="CRJ6B037V2ZT4U3W17VC",
    ),
    rgw: RGWAdmin = Depends(get_rgw_admin),
    user: UserDB = Depends(get_user_by_path_uid),
) -> None:
    """
    Delete a specific S3 Access key for a specific user.
    \f
    Parameters
    ----------
    access_id : str
        ID of the S3 key to delete. URL Path Parameter.
    rgw : rgwadmin.RGWAdmin
        RGW admin interface to manage Ceph's object store. Dependency Injection.
    user : app.models.user.User
        User with given uid. Dependency Injection.
    """
    if len(rgw.get_user(uid=user.uid, stats=False)["keys"]) <= 1:
        raise HTTPException(status.HTTP_400_BAD_REQUEST, detail="It's not possible to delete the last key")
    try:
        rgw.remove_key(access_key=access_id, uid=user.uid)
    except RGWAdminException:
        raise HTTPException(status.HTTP_404_NOT_FOUND, detail="Key not found")
