import secrets
from typing import Any, Dict, List, Optional, Union

from pydantic import AnyHttpUrl, AnyUrl, BaseSettings, Field, validator


def _assemble_db_uri(values: Dict[str, Any], async_flag: bool = True) -> Any:
    return AnyUrl.build(
        scheme=f"mysql+{'aiomysql' if async_flag else 'pymysql'}",
        password=values.get("DB_PASSWORD"),
        user=values.get("DB_USER"),
        port=str(values.get("DB_PORT")),
        host=values.get("DB_HOST"),
        path=f"/{values.get('DB_DATABASE') or ''}",
    )


class Settings(BaseSettings):
    DOMAIN: str = Field("localhost", description="Domain of the service.")
    SSL_TERMINATION: bool = Field(False, description="Flag if the service runs behind a SSL termination proxy")
    API_PREFIX: str = Field("/api", description="Path Prefix for all API endpoints.")
    SECRET_KEY: str = Field(secrets.token_urlsafe(32), description="Secret key to sign the JWTs.")
    # 60 minutes * 24 hours * 8 days = 8 days
    JWT_TOKEN_EXPIRE_MINUTES: int = Field(60 * 24 * 8, description="JWT lifespan in minutes.")
    # BACKEND_CORS_ORIGINS is a JSON-formatted list of origins
    BACKEND_CORS_ORIGINS: List[AnyHttpUrl] = Field([], description="List of all valid CORS origins")

    @validator("BACKEND_CORS_ORIGINS", pre=True)
    def assemble_cors_origins(cls, v: Union[str, List[str]]) -> Union[List[str], str]:
        if isinstance(v, str) and not v.startswith("["):
            return [i.strip() for i in v.split(",")]
        elif isinstance(v, (list, str)):
            return v
        raise ValueError(v)

    DB_HOST: str = Field(..., description="Host of the database.")
    DB_USER: str = Field(..., description="Username in the database.")
    DB_PASSWORD: str = Field(..., description="Password for the database user.")
    DB_DATABASE: str = Field(..., description="Name of the database.")
    DB_PORT: int = Field(3306, description="Port of the database.")
    SQLALCHEMY_VERBOSE_LOGGER: bool = Field(False, description="Flag whether to print the SQL Queries in the logs.")
    SQLALCHEMY_DATABASE_ASYNC_URI: AnyUrl | None = None

    @validator("SQLALCHEMY_DATABASE_ASYNC_URI", pre=True)
    def assemble_async_db_connection(cls, v: Optional[str], values: Dict[str, Any]) -> Any:
        if isinstance(v, str):
            return v
        return _assemble_db_uri(values, async_flag=True)

    SQLALCHEMY_DATABASE_NORMAL_URI: AnyUrl | None = None

    @validator("SQLALCHEMY_DATABASE_NORMAL_URI", pre=True)
    def assemble_db_connection(cls, v: Optional[str], values: Dict[str, Any]) -> Any:
        if isinstance(v, str):
            return v
        return _assemble_db_uri(values, async_flag=False)

    OBJECT_GATEWAY_URI: AnyHttpUrl = Field(..., description="URI of the Ceph Object Gateway.")
    CEPH_ACCESS_KEY: str = Field(..., description="Access key for the Ceph Object Gateway with admin privileges.")
    CEPH_SECRET_KEY: str = Field(..., description="Secret key for the Ceph Object Gateway with admin privileges.")
    CEPH_USERNAME: str = Field(..., description="ID of the Proxy user in Ceph.")

    OIDC_CLIENT_SECRET: str = Field(..., description="OIDC Client secret")
    OIDC_CLIENT_ID: str = Field(..., description="OIDC Client ID")
    OIDC_BASE_URI: AnyHttpUrl = Field(..., description="OIDC Base URI")
    OIDC_META_INFO_PATH: str = Field("/.well-known/openid-configuration", description="Path to the OIDC meta data file")

    class Config:
        case_sensitive = True
        env_file = ".env"
        secrets_dir = "/run/secrets"


settings = Settings()
