from datetime import datetime, timedelta
from typing import Any, Union

from authlib.integrations.starlette_client import OAuth
from authlib.jose import JsonWebToken

from app.core.config import settings

ISSUER = "proxy_api"
ALGORITHM = "HS256"
jwt = JsonWebToken([ALGORITHM])


def create_access_token(subject: Union[str, Any]) -> str:
    """
    Create a JWT access token.

    Parameters
    ----------
    subject : Union[str, Any]
        Ths subject in the JWT.

    Returns
    -------
    token : str
        The generated JWT.

    """
    expire = datetime.utcnow() + timedelta(minutes=settings.JWT_TOKEN_EXPIRE_MINUTES)
    to_encode = {"exp": expire, "sub": str(subject), "iss": ISSUER}
    encoded_jwt = jwt.encode(header={"alg": ALGORITHM}, payload=to_encode, key=settings.SECRET_KEY)
    return encoded_jwt.decode("utf-8")


def decode_token(token: str) -> dict[str, str]:
    """
    Decode and verify a JWT token.

    Parameters
    ----------
    token : str
        The JWT to decode.

    Returns
    -------
    decoded_token : dict[str, str]
        Payload of the decoded token.
    """
    claims = jwt.decode(
        s=token,
        key=settings.SECRET_KEY,
        claims_options={
            "iss": {"essential": True},
            "sub": {"essential": True},
            "exp": {"essential": True},
        },
    )
    claims.validate()
    return claims


oauth = OAuth()
oauth.register(
    name="lifescience",
    client_id=settings.OIDC_CLIENT_ID,
    client_secret=settings.OIDC_CLIENT_SECRET,
    server_metadata_url=settings.OIDC_BASE_URI + settings.OIDC_META_INFO_PATH,
    client_kwargs={"scope": "openid profile aarc", "code_challenge_method": "S256"},
)
