from sqlalchemy import func, or_
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import joinedload

from app.models.bucket import Bucket
from app.models.bucket_permission import BucketPermission as BucketPermissionDB
from app.models.bucket_permission import PermissionEnum
from app.schemas.bucket import BucketIn as BucketInSchema


class CRUDBucket:
    @staticmethod
    async def get(db: AsyncSession, bucket_name: str) -> Bucket | None:
        """
        Get a bucket by its name.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        bucket_name : str
            Name of the Bucket to get from database.

        Returns
        -------
        bucket : Bucket | None
            Returns the bucket if it exists, None otherwise.
        """
        stmt = select(Bucket).where(Bucket.name == bucket_name)
        row = await db.execute(stmt)
        return row.scalar()

    @staticmethod
    async def get_for_user(db: AsyncSession, uid: str) -> list[Bucket]:
        """
        Get all buckets where the given user has READ permissions for.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        uid : str
            UID of a user.

        Returns
        -------
        buckets : list[app.models.bucket.Bucket]
            A list of all buckets where the given user has READ permissions for.

        Notes
        -----
        Creates this SQL Query
                SELECT bucket.name, bucket.description, bucket.public, bucket.owner_id,
                       user_1.uid, user_1.display_name
                FROM bucket LEFT OUTER JOIN user AS user_1 ON user_1.uid = bucket.owner_id
                WHERE bucket.owner_id = %s OR (EXISTS
                    (SELECT 1 FROM bucketpermission
                    WHERE bucket.name = bucketpermission.bucket_name AND bucketpermission.user_id = %s
                        AND(bucketpermission.permissions = %s OR bucketpermission.permissions = %s)
                        AND(datediff(now(), bucketpermission.`from`) <= %s OR bucketpermission.`from` IS NULL)
                        AND(datediff(now(), bucketpermission.`to`) >= %s OR bucketpermission.`to` IS NULL)))
        """
        stmt = (
            select(Bucket)
            .options(joinedload(Bucket.owner))
            .where(
                or_(
                    Bucket.owner_id == uid,
                    Bucket.permissions.any(BucketPermissionDB.user_id == uid)
                    .where(
                        or_(
                            BucketPermissionDB.permissions == PermissionEnum.READ,
                            BucketPermissionDB.permissions == PermissionEnum.READWRITE,
                        )
                    )
                    .where(
                        or_(
                            func.datediff(func.now(), BucketPermissionDB.from_) >= 0,
                            BucketPermissionDB.from_ == None,  # noqa:E711
                        )
                    )
                    .where(
                        or_(
                            func.datediff(func.now(), BucketPermissionDB.to) <= 0,
                            BucketPermissionDB.to == None,  # noqa:E711
                        )
                    ),
                )
            )
        )

        buckets = (await db.execute(stmt)).scalars().all()
        return buckets

    @staticmethod
    async def get_own_buckets(db: AsyncSession, uid: str) -> list[Bucket]:
        """
        Get all the buckets where the user is the owner of it.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        uid : str
            UID of the user to get the buckets for.

        Returns
        -------
        buckets : list[app.models.bucket.Bucket]
            All the buckets for the given UID.
        """
        stmt = select(Bucket).where(Bucket.owner_id == uid)
        return (await db.execute(stmt)).scalars().all()

    @staticmethod
    async def create(db: AsyncSession, bucket_in: BucketInSchema, uid: str) -> Bucket | None:
        """
        Create a bucket for a given user.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        bucket_in : app.schemas.bucket.BucketIn
            All relevant information for a new bucket.
        uid : str
            UID of the owner for the new bucket.

        Returns
        -------
        bucket : app.models.bucket.Bucket | None
            Returns the created bucket. If None then there was a problem, e.g. the name of the bucket is already taken.
        """
        bucket = Bucket(**bucket_in.dict(), owner_id=uid)
        if await CRUDBucket.get(db, bucket.name) is None:
            db.add(bucket)
            await db.commit()
            await db.refresh(bucket)
            return bucket
        return None

    @staticmethod
    async def delete(db: AsyncSession, bucket: Bucket) -> None:
        """
        Delete a specific bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        bucket : app.models.bucket.Bucket
            The bucket to delete.
        """
        await db.delete(bucket)
        await db.commit()
