from sqlalchemy import and_
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import joinedload

from app.crud.crud_bucket import CRUDBucket
from app.crud.crud_user import CRUDUser
from app.models.bucket_permission import BucketPermission as BucketPermissionDB
from app.schemas.bucket_permission import BucketPermissionIn as BucketPermissionSchema
from app.schemas.bucket_permission import BucketPermissionParameters as BucketPermissionParametersSchema


class CRUDBucketPermission:
    @staticmethod
    async def get(db: AsyncSession, bucket_name: str, user_id: str) -> BucketPermissionDB | None:
        stmt = select(BucketPermissionDB).where(
            and_(BucketPermissionDB.user_id == user_id, BucketPermissionDB.bucket_name == bucket_name)
        )
        row = await db.execute(stmt)
        return row.scalar()

    @staticmethod
    async def get_permissions_for_bucket(db: AsyncSession, bucket_name: str) -> list[BucketPermissionDB]:
        """
        Get the permissions for the given bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        bucket_name : str
            Name of the bucket which to query.

        Returns
        -------
        buckets : list[BucketPermission]
            Returns the permissions for the given bucket.
        """
        stmt = (
            select(BucketPermissionDB)
            .options(joinedload(BucketPermissionDB.grantee))
            .where(BucketPermissionDB.bucket_name == bucket_name)
        )
        row = await db.execute(stmt)
        return row.scalars().all()

    @staticmethod
    async def get_permissions_for_user(db: AsyncSession, user_id: str) -> list[BucketPermissionDB]:
        """
        Get the permissions for the given user.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        user_id : str
            UID of the user which to query.

        Returns
        -------
        buckets : list[BucketPermission]
            Returns the permissions for the given user.
        """
        stmt = select(BucketPermissionDB).where(BucketPermissionDB.user_id == user_id)
        row = await db.execute(stmt)
        return row.scalars().all()

    @staticmethod
    async def check_permission(db: AsyncSession, bucket_name: str, uid: str, only_own: bool = False) -> bool:
        """
        Check if the provided user has READ permission to the provided bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        bucket_name : str
            Name of the bucket for which to perform the check.
        uid : str
            UID of the user for which to perform the check.
        only_own : bool, default False
            Flag if the check is only for the users own buckets or include foreign buckets with permission.

        Returns
        -------
        permission_check : bool
            Return True if the user has READ permission on the bucket, False otherwise.
        """
        buckets = await (CRUDBucket.get_own_buckets(db, uid) if only_own else CRUDBucket.get_for_user(db, uid))
        return bucket_name in map(lambda x: x.name, buckets)

    @staticmethod
    async def create(db: AsyncSession, permission: BucketPermissionSchema) -> BucketPermissionDB:
        """
        Create a permission in the database and raise Exceptions if there are problems.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        permission : app.schemas.bucket_permission.BucketPermissionOut
            The permission to create.
        Returns
        -------
        permission : app.models.bucket_permission.BucketPermission
            Newly created permission model from the db.
        """
        # Check if user exists
        user = await CRUDUser.get(db, uid=permission.uid)
        if user is None:
            raise KeyError(
                f"Unknown user with uid {permission.uid}",
            )
        # Check that grantee is not the owner of the bucket
        bucket = await CRUDBucket.get(db, permission.bucket_name)
        if bucket is None or bucket.owner_id == user.uid:
            raise ValueError(f"User {permission.uid} is the owner of the bucket {permission.bucket_name}")
        # Check if combination of user and bucket already exists
        duplicate_check_stmt = select(BucketPermissionDB).where(
            and_(BucketPermissionDB.user_id == user.uid, BucketPermissionDB.bucket_name == permission.bucket_name)
        )
        previous_permission = (await db.execute(duplicate_check_stmt)).scalar()
        if previous_permission is not None:
            raise DuplicateError(
                f"bucket permission for combination {permission.bucket_name} {permission.uid} already exists."
            )
        # Add permission to db
        permission_db = BucketPermissionDB(
            user_id=user.uid,
            bucket_name=permission.bucket_name,
            from_=permission.from_timestamp,
            to=permission.to_timestamp,
            file_prefix=permission.file_prefix,
            permissions=permission.permission,
        )
        db.add(permission_db)
        await db.commit()
        await db.refresh(permission_db)
        return permission_db

    @staticmethod
    async def delete(db: AsyncSession, permission: BucketPermissionDB) -> None:
        """
        Delete a permission in the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        permission : app.schemas.bucket_permission.BucketPermissionOut
            The permission to create.
        Returns
        -------
        permission : app.models.bucket_permission.BucketPermission
            Newly created permission model from the db.
        """
        await db.delete(permission)
        await db.commit()

    @staticmethod
    async def update_permission(
        db: AsyncSession, permission: BucketPermissionDB, new_params: BucketPermissionParametersSchema
    ) -> BucketPermissionDB:
        """
        Update a permission in the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession
            Async database session to perform query on.
        permission : app.schemas.bucket_permission.BucketPermissionOut
            The permission to update.
        new_params : app.schemas.bucket_permission.BucketPermissionParameters
            The parameters which should be updated.

        Returns
        -------
        permission : app.models.bucket_permission.BucketPermission
            Updated permission model from the db.
        """
        permission.update_parameters(new_params)
        await db.commit()
        await db.refresh(permission)
        return permission


class DuplicateError(Exception):
    pass
