from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from app.models.user import User


class CRUDUser:
    @staticmethod
    async def create(db: AsyncSession, user: User) -> User:
        """
        Create a new user in the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        user : app.models.user.User
            The user to create.

        Returns
        -------
        user : app.models.user.User
            The newly created user.
        """
        db.add(user)
        await db.commit()
        await db.refresh(user)
        return user

    @staticmethod
    async def get(db: AsyncSession, uid: str) -> User | None:
        """
        Get a user by its UID.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        uid : str
            UID of a user.

        Returns
        -------
        user : app.models.user.User | None
            The user for the given UID if he exists, None otherwise
        """
        stmt = select(User).where(User.uid == uid)
        return (await db.execute(stmt)).scalar()

    @staticmethod
    async def search_for_name(db: AsyncSession, name_substring: str) -> list[User]:
        """
        Search for users that contain a specific substring in their name.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        name_substring : str
            Substring to search for in the name of a user.

        Returns
        -------
        users : list[app.models.user.User]
            List of users which have the given substring in their name.
        """
        stmt = select(User).where(User.display_name.contains(name_substring))
        return (await db.execute(stmt)).scalars().all()
