import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import RedirectResponse
from fastapi.routing import APIRoute
from starlette.middleware.sessions import SessionMiddleware

from app.api.api import api_router
from app.api.endpoints.login import LoginException, login_exception_handler
from app.api.miscellaneous_endpoints import miscellaneous_router
from app.core.config import settings

description = """
This is the backend for a new UI which can leverage the additional powerful functionality provided by Ceph in a
user-friendly manner.
"""


def custom_generate_unique_id(route: APIRoute) -> str:
    return f"{route.tags[-1]}-{route.name}"


app = FastAPI(
    title="S3-Proxy",
    version="1.0.0",
    description=description,
    contact={
        "name": "Daniel Goebel",
        "url": "https://ekvv.uni-bielefeld.de/pers_publ/publ/PersonDetail.jsp?personId=223066601",
        "email": "dgoebel@techfak.uni-bielefeld.de",
    },
    generate_unique_id_function=custom_generate_unique_id,
    # license_info={"name": "MIT", "url": "https://mit-license.org/"},
    root_path=settings.API_PREFIX,
)

# CORS Settings for the API
app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.BACKEND_CORS_ORIGINS,
    allow_credentials=False,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Enable gzip compression for large responses
app.add_middleware(GZipMiddleware, minimum_size=500)

# Include all routes
app.include_router(api_router)
app.include_router(miscellaneous_router)
app.add_exception_handler(LoginException, login_exception_handler)

app.add_middleware(SessionMiddleware, secret_key=settings.SECRET_KEY)


@app.get("/", response_class=RedirectResponse, tags=["Miscellaneous"], include_in_schema=False)
def redirect_docs() -> str:
    return settings.API_PREFIX + "/docs"


if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8000)
