from datetime import datetime
from enum import Enum, unique
from typing import TYPE_CHECKING

from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.dialects.mysql import ENUM, TIMESTAMP
from sqlalchemy.orm import relationship

from app.db.base_class import Base

if TYPE_CHECKING:
    from app.schemas.bucket_permission import BucketPermissionParameters

    from .bucket import Bucket
    from .user import User
else:
    BucketPermissionParameters = object


@unique
class PermissionEnum(str, Enum):
    """
    Enumeration for the possible permission on a bucket.
    """

    READ: str = "READ"
    WRITE: str = "WRITE"
    READWRITE: str = "READWRITE"


class BucketPermission(Base):
    """
    Database model for the permission for a user on a bucket.
    Will be deleted if either the user or the bucket is deleted.
    """

    __tablename__: str = "bucketpermission"
    user_id: str = Column(ForeignKey("user.uid", ondelete="CASCADE"), primary_key=True)
    bucket_name: str = Column(ForeignKey("bucket.name", ondelete="CASCADE"), primary_key=True)
    from_: datetime | None = Column("from", TIMESTAMP, nullable=True)
    to: datetime | None = Column(TIMESTAMP, nullable=True)
    file_prefix: str | None = Column(String(512), nullable=True)
    permissions: str | PermissionEnum = Column(ENUM(PermissionEnum), default=PermissionEnum.READ, nullable=False)
    grantee: "User" = relationship("User", back_populates="permissions")
    bucket: "Bucket" = relationship("Bucket", back_populates="permissions")

    def update_parameters(self, params: BucketPermissionParameters) -> None:  # pragma: no cover
        """
        Update the object with the new parameters.

        Parameters
        ----------
        params : app.schemas.bucket_permission.BucketPermissionParameters
            The parameters which should be updated.
        """
        self.from_ = params.from_timestamp
        self.to = params.to_timestamp
        self.file_prefix = params.file_prefix
        self.permissions = params.permission

    def __repr__(self) -> str:
        return f"BucketPermission(uid={self.user_id} bucket_name={self.bucket_name})"
