from datetime import datetime
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from mypy_boto3_s3.service_resource import ObjectSummary
else:
    ObjectSummary = object


class _BaseBucket(BaseModel):
    """
    Base Schema for a bucket.
    """

    name: str = Field(
        ...,
        example="test-bucket",
        description="Name of the bucket",
        min_length=3,
        max_length=63,
        regex=r"(?!(^((2(5[0-5]|[0-4]\d)|[01]?\d{1,2})\.){3}(2(5[0-5]|[0-4]\d)|[01]?\d{1,2})$))^[a-z\d][a-z\d.-]{1,61}[a-z\d]$",  # noqa:E501
    )
    description: str = Field(
        ...,
        example="""\
This is a very long sample description of a bucket and its purpose which has to be more \
than 126 characters long and is mandatory.
                             """.strip(),
        description="Description of the bucket",
        min_length=126,
        max_length=2**16,
    )


class BucketIn(_BaseBucket):
    """
    Schema for creating a new bucket.
    """


class BucketOut(_BaseBucket):
    """
    Schema for answering a request with a bucket.
    """

    created_at: datetime = Field(
        ...,
        example=datetime(2022, 1, 1, 0, 0),
        description="Time when the bucket was created",
    )
    owner: str = Field(..., description="UID of the owner", example="28c5353b8bb34984a8bd4169ba94c606")
    num_objects: int = Field(..., description="Number of Objects in this bucket", example=6)
    size: int = Field(..., description="Total size of objects in this bucket in bytes", example=3256216)

    class Config:
        orm_mode = True


class S3ObjectMetaInformation(BaseModel):
    """
    Schema for the meta-information about a S3 object.
    """

    key: str = Field(
        ...,
        description="Key of the Object in the S3 store",
        example="test.txt",
        max_length=512,
    )
    bucket: str = Field(
        ...,
        description="Name of the Bucket in which the object is",
        example="test-bucket",
        max_length=256,
    )
    content_type: str = Field(..., description="MIME type of the object", example="text/plain")
    size: int = Field(..., description="Size of the object in Bytes", example=123456)
    last_modified: datetime = Field(
        ...,
        description="Last time the object was modified",
        example=datetime(2022, 1, 1, 0, 0),
    )

    @staticmethod
    def from_native_s3_object(obj: ObjectSummary) -> "S3ObjectMetaInformation":  # pragma: no cover
        """
        Create an S3ObjectMetaInformation object from a boto3 S3ObjectSummary object.

        Parameters
        ----------
        obj : mypy_boto3_s3.service_resource.ObjectSummary
            boto3 S3ObjectSummary.

        Returns
        -------
        obj : app.schemas.bucket.S3ObjectMetaInformation
            Converted S3objectMetaInformation.
        """
        return S3ObjectMetaInformation(
            key=obj.key,
            bucket=obj.bucket_name,
            size=obj.size,
            last_modified=obj.last_modified,
            content_type=obj.Object().content_type,
        )
