import json
from datetime import datetime, timedelta

import pytest
from fastapi import status
from httpx import AsyncClient
from sqlalchemy.ext.asyncio import AsyncSession

from app.models.bucket import Bucket
from app.models.bucket_permission import PermissionEnum
from app.models.user import User
from app.schemas.bucket_permission import BucketPermissionIn as BucketPermissionSchema
from app.schemas.bucket_permission import BucketPermissionParameters as BucketPermissionParametersSchema
from app.tests.utils.bucket import add_permission_for_bucket
from app.tests.utils.user import get_authorization_headers
from app.tests.utils.utils import json_datetime_converter


class _TestBucketPermissionRoutes:
    base_path = "/permissions/"


class TestBucketPermissionRoutesGet(_TestBucketPermissionRoutes):
    @pytest.mark.asyncio
    async def test_get_valid_bucket_permission(
        self,
        client: AsyncClient,
        user_token_headers: dict[str, str],
        random_bucket_permission_schema: BucketPermissionSchema,
    ) -> None:
        """
        Test for getting a valid bucket permission.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        response = await client.get(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/{random_bucket_permission_schema.uid}",  # noqa:E501
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_200_OK

        permission = response.json()

        assert permission
        assert permission["uid"] == random_bucket_permission_schema.uid
        assert permission["bucket_name"] == random_bucket_permission_schema.bucket_name

    @pytest.mark.asyncio
    async def test_get_bucket_permission_for_unknown_user(
        self,
        client: AsyncClient,
        user_token_headers: dict[str, str],
        random_bucket_permission_schema: BucketPermissionSchema,
    ) -> None:
        """
        Test for getting a bucket permission for an unknown user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        response = await client.get(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/ImpossibleUser",
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_get_unknown_bucket_permission(
        self, client: AsyncClient, user_token_headers: dict[str, str], random_bucket: Bucket, random_second_user: User
    ) -> None:
        """
        Test for getting a bucket permission for an unknown user/bucket combination.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        response = await client.get(
            f"{self.base_path}bucket/{random_bucket.name}/user/{random_second_user.uid}",
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_get_foreign_bucket_permission_with_permission(
        self, client: AsyncClient, random_bucket_permission_schema: BucketPermissionSchema, random_second_user: User
    ) -> None:
        """
        Test for getting a bucket permission for a foreign bucket with READ permission for that bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        user_token_headers = get_authorization_headers(random_second_user.uid)
        response = await client.get(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/{random_second_user.uid}",
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_200_OK
        permission = response.json()

        assert permission
        assert permission["uid"] == random_bucket_permission_schema.uid
        assert permission["bucket_name"] == random_bucket_permission_schema.bucket_name

    @pytest.mark.asyncio
    async def test_get_wrong_bucket_permission_with_permission(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_bucket_permission_schema: BucketPermissionSchema,
        random_third_user: User,
    ) -> None:
        """
        Test for getting a bucket permission as a grantee for another grantee for the bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        random_third_user : app.models.user.User
            Random third user who has no permissions for the bucket. pytest fixture.
        """
        await add_permission_for_bucket(db, random_bucket_permission_schema.bucket_name, random_third_user.uid)
        user_token_headers = get_authorization_headers(random_third_user.uid)
        response = await client.get(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/{random_bucket_permission_schema.uid}",  # noqa:E501
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.asyncio
    async def test_get_bucket_permissions_for_user(
        self, client: AsyncClient, random_second_user: User, random_bucket_permission_schema: BucketPermissionSchema
    ) -> None:
        """
        Test for getting all bucket permission for a user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        user_token_headers = get_authorization_headers(random_second_user.uid)
        response = await client.get(
            f"{self.base_path}user/{random_bucket_permission_schema.uid}",
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_200_OK
        permission_list = response.json()
        assert isinstance(permission_list, list)
        assert len(permission_list) == 1
        permission = permission_list[0]
        assert permission["uid"] == random_bucket_permission_schema.uid
        assert permission["bucket_name"] == random_bucket_permission_schema.bucket_name

    @pytest.mark.asyncio
    async def test_get_bucket_permissions_for_bucket(
        self,
        client: AsyncClient,
        user_token_headers: dict[str, str],
        random_bucket_permission_schema: BucketPermissionSchema,
    ) -> None:
        """
        Test for getting all bucket permission for a bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        response = await client.get(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}",
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_200_OK
        permission_list = response.json()
        assert isinstance(permission_list, list)
        assert len(permission_list) == 1
        permission = permission_list[0]
        assert permission["uid"] == random_bucket_permission_schema.uid
        assert permission["bucket_name"] == random_bucket_permission_schema.bucket_name

    @pytest.mark.asyncio
    async def test_get_bucket_permissions_for_foreign_bucket(
        self, client: AsyncClient, random_second_user: User, random_bucket_permission_schema: BucketPermissionSchema
    ) -> None:
        """
        Test for getting all bucket permissions for a foreign bucket with READ permission for that bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        user_token_headers = get_authorization_headers(random_second_user.uid)
        response = await client.get(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}",
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_403_FORBIDDEN


class TestBucketPermissionRoutesCreate(_TestBucketPermissionRoutes):
    @pytest.mark.asyncio
    async def test_create_bucket_permissions_for_unknown_user(
        self, client: AsyncClient, user_token_headers: dict[str, str], random_bucket: Bucket
    ) -> None:
        """
        Test for creating a bucket permission for an unknown user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        permission = BucketPermissionSchema(bucket_name=random_bucket.name, uid="ImpossibleUser")
        response = await client.post(self.base_path, headers=user_token_headers, json=permission.dict())
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_create_bucket_permissions_for_owner(
        self, client: AsyncClient, user_token_headers: dict[str, str], random_user: User, random_bucket: Bucket
    ) -> None:
        """
        Test for creating a bucket permission for the owner of the bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_user : app.models.user.User
            Random user for testing who is owner of the bucket. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        permission = BucketPermissionSchema(bucket_name=random_bucket.name, uid=random_user.uid)
        response = await client.post(self.base_path, headers=user_token_headers, json=permission.dict())
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_create_duplicate_bucket_permissions(
        self,
        client: AsyncClient,
        user_token_headers: dict[str, str],
        random_bucket_permission_schema: BucketPermissionSchema,
    ) -> None:
        """
        Test for creating a duplicated bucket permission.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        permission = BucketPermissionSchema(
            bucket_name=random_bucket_permission_schema.bucket_name, uid=random_bucket_permission_schema.uid
        )
        response = await client.post(self.base_path, headers=user_token_headers, json=permission.dict())
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_create_valid_bucket_permissions(
        self, client: AsyncClient, user_token_headers: dict[str, str], random_second_user: User, random_bucket: Bucket
    ) -> None:
        """
        Test for creating a valid bucket permission.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        permission = BucketPermissionSchema(bucket_name=random_bucket.name, uid=random_second_user.uid)

        response = await client.post(self.base_path, headers=user_token_headers, json=permission.dict())
        assert response.status_code == status.HTTP_201_CREATED
        created_permission = response.json()
        assert created_permission["uid"] == random_second_user.uid
        assert created_permission["bucket_name"] == random_bucket.name
        assert created_permission["grantee_display_name"] == random_second_user.display_name


class TestBucketPermissionRoutesDelete(_TestBucketPermissionRoutes):
    @pytest.mark.asyncio
    async def test_delete_bucket_permission_from_owner(
        self,
        client: AsyncClient,
        user_token_headers: dict[str, str],
        random_bucket_permission_schema: BucketPermissionSchema,
    ) -> None:
        """
        Test for deleting a valid bucket permission as the owner of the bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        response = await client.delete(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/{random_bucket_permission_schema.uid}",  # noqa:E501
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_204_NO_CONTENT

    @pytest.mark.asyncio
    async def test_delete_foreign_bucket_permission_with_permission(
        self, client: AsyncClient, random_second_user: User, random_bucket_permission_schema: BucketPermissionSchema
    ) -> None:
        """
        Test for deleting a bucket permission as a grantee.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        user_token_headers = get_authorization_headers(random_second_user.uid)
        response = await client.get(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/{random_bucket_permission_schema.uid}",  # noqa:E501
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.asyncio
    async def test_delete_bucket_permission_with_unknown_user(
        self,
        client: AsyncClient,
        user_token_headers: dict[str, str],
        random_bucket_permission_schema: BucketPermissionSchema,
    ) -> None:
        """
        Test for deleting a bucket permission as the grantee of the permission.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        response = await client.delete(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/ImpossibleUser",
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_delete_bucket_permission_without_permission(
        self, client: AsyncClient, user_token_headers: dict[str, str], random_bucket: Bucket, random_second_user: User
    ) -> None:
        """
        Test for deleting a bucket permission with an unknown bucket/user combination.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        response = await client.delete(
            f"{self.base_path}bucket/{random_bucket.name}/user/{random_second_user.uid}",
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_delete_wrong_bucket_permission_with_permission(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_bucket_permission_schema: BucketPermissionSchema,
        random_third_user: User,
    ) -> None:
        """
        Test for deleting a bucket permission as a grantee for another grantee for the bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        random_third_user : app.models.user.User
            Random third user who has no permissions for the bucket. pytest fixture.
        """
        await add_permission_for_bucket(db, random_bucket_permission_schema.bucket_name, random_third_user.uid)
        user_token_headers = get_authorization_headers(random_third_user.uid)
        response = await client.delete(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/{random_bucket_permission_schema.uid}",  # noqa:E501
            headers=user_token_headers,
        )
        assert response.status_code == status.HTTP_403_FORBIDDEN


class TestBucketPermissionRoutesUpdate(_TestBucketPermissionRoutes):
    @pytest.mark.asyncio
    async def test_update_valid_bucket_permission(
        self,
        client: AsyncClient,
        user_token_headers: dict[str, str],
        random_bucket_permission_schema: BucketPermissionSchema,
    ) -> None:
        """
        Test for updating a bucket permission as the owner of the bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        new_from_time = datetime(2022, 1, 1, 0, 0)
        new_params = BucketPermissionParametersSchema(
            from_timestamp=new_from_time,
            to_timestamp=new_from_time + timedelta(days=1),
            permission=PermissionEnum.READWRITE,
            file_prefix="pseudo/folder/",
        )
        response = await client.put(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/{random_bucket_permission_schema.uid}",  # noqa:E501
            headers=user_token_headers,
            content=json.dumps(new_params.dict(), default=json_datetime_converter),
        )
        assert response.status_code == status.HTTP_200_OK
        updated_permission = response.json()
        assert updated_permission["uid"] == random_bucket_permission_schema.uid
        assert updated_permission["bucket_name"] == random_bucket_permission_schema.bucket_name
        if new_params.from_timestamp is not None and new_params.to_timestamp is not None:
            assert updated_permission["from_timestamp"] == new_params.from_timestamp.strftime("%Y-%m-%dT%H:%M:%S")
            assert updated_permission["to_timestamp"] == new_params.to_timestamp.strftime("%Y-%m-%dT%H:%M:%S")
        assert updated_permission["permission"] == new_params.permission
        assert updated_permission["file_prefix"] == new_params.file_prefix

    @pytest.mark.asyncio
    async def test_update_unknown_bucket_permission(
        self,
        client: AsyncClient,
        user_token_headers: dict[str, str],
        random_bucket_permission_schema: BucketPermissionSchema,
    ) -> None:
        """
        Test for updating a bucket permission with an unknown user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        new_params = BucketPermissionParametersSchema(
            permission=PermissionEnum.READWRITE,
            file_prefix="pseudo/folder/",
        )
        response = await client.put(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/impossibleUser",
            headers=user_token_headers,
            json=new_params.dict(),
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_update_bucket_permission_without_permission(
        self, client: AsyncClient, user_token_headers: dict[str, str], random_bucket: Bucket, random_second_user: User
    ) -> None:
        """
        Test for updating a non-existing bucket permission with a valid user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        new_params = BucketPermissionParametersSchema(
            permission=PermissionEnum.READWRITE,
            file_prefix="pseudo/folder/",
        )
        response = await client.put(
            f"{self.base_path}bucket/{random_bucket.name}/user/{random_second_user.uid}",
            headers=user_token_headers,
            json=new_params.dict(),
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_update_foreign_bucket_permission_with_permission(
        self, client: AsyncClient, random_bucket_permission_schema: BucketPermissionSchema, random_second_user: User
    ) -> None:
        """
        Test for updating a bucket permission as the grantee of the permission.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        user_token_headers = get_authorization_headers(random_second_user.uid)
        new_params = BucketPermissionParametersSchema(
            permission=PermissionEnum.READWRITE,
            file_prefix="pseudo/folder/",
        )
        response = await client.put(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/{random_second_user.uid}",
            headers=user_token_headers,
            json=new_params.dict(),
        )
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.asyncio
    async def test_update_foreign_bucket_permission_without_permission(
        self, client: AsyncClient, random_bucket_permission_schema: BucketPermissionSchema, random_third_user: User
    ) -> None:
        """
        Test for updating a bucket permission as an unrelated third user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        random_third_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        user_token_headers = get_authorization_headers(random_third_user.uid)
        new_params = BucketPermissionParametersSchema(
            permission=PermissionEnum.READWRITE,
            file_prefix="pseudo/folder/",
        )
        response = await client.put(
            f"{self.base_path}bucket/{random_bucket_permission_schema.bucket_name}/user/{random_bucket_permission_schema.uid}",  # noqa:E501
            headers=user_token_headers,
            json=new_params.dict(),
        )
        assert response.status_code == status.HTTP_403_FORBIDDEN
