from urllib.parse import parse_qs, urlparse

import pytest
from fastapi import status
from httpx import AsyncClient
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from app.core.config import settings
from app.core.security import decode_token
from app.models.user import User
from app.tests.mocks.mock_rgw_admin import MockRGWAdmin
from app.tests.utils.utils import random_lower_string


class TestLoginRoute:
    login_path: str = "/auth/"

    @pytest.mark.asyncio
    async def test_login_redirect(self, client: AsyncClient) -> None:
        """
        Test for the query parameter on the login redirect route.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        """
        r = await client.get(self.login_path + "login", follow_redirects=False)
        query = parse_qs(urlparse(r.headers["location"], scheme="http").query)
        assert r.status_code == status.HTTP_302_FOUND
        assert "code" in query["response_type"]
        assert settings.OIDC_CLIENT_ID in query["client_id"]
        assert "openid" in query["scope"][0].split(" ")
        assert query["state"]

    @pytest.mark.asyncio
    async def test_successful_login_with_existing_user(self, client: AsyncClient, random_user: User) -> None:
        """
        Test for login callback route with an existing user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_user : app.models.user.User
            Random user for testing. pytest fixture.
        """
        r = await client.get(
            self.login_path + "callback",
            params={
                "voperson_id": random_user.uid,
                "name": random_user.display_name,
            },
            follow_redirects=False,
        )
        assert r.status_code == status.HTTP_302_FOUND
        assert "set-cookie" in r.headers.keys()
        cookie_header = r.headers["set-cookie"]
        right_header = None
        for t in cookie_header.split(";"):
            if t.startswith("bearer"):
                right_header = t
                break
        assert right_header
        claim = decode_token(right_header.split("=")[1])
        assert claim["sub"] == random_user.uid

    @pytest.mark.asyncio
    async def test_login_with_error(self, client: AsyncClient) -> None:
        """
        Test for login callback route with an existing user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        """
        r = await client.get(
            self.login_path + "callback",
            params={"voperson_id": "", "name": "", "error": True},
            follow_redirects=False,
        )
        assert r.status_code == status.HTTP_302_FOUND
        assert "set-cookie" in r.headers.keys()
        assert find_cookie(searched_cookie_name="bearer", cookie_header=r.headers["set-cookie"]) is None
        assert r.headers["location"].startswith("/?login_error")

    @pytest.mark.asyncio
    async def test_successful_login_with_non_existing_user(
        self, client: AsyncClient, mock_rgw_admin: MockRGWAdmin, db: AsyncSession
    ) -> None:
        """
        Test for login callback route with a non-existing user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        mock_rgw_admin : app.tests.mocks.mock_rgw_admin.MockRGWAdmin
            Mock RGW admin for Ceph. pytest fixture.
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        """
        uid = random_lower_string()
        display_name = f"{random_lower_string(8)} {random_lower_string(8)}"
        r = await client.get(
            self.login_path + "callback",
            params={"voperson_id": uid, "name": display_name},
            follow_redirects=False,
        )
        # Check response and valid/right jwt token
        assert r.status_code == status.HTTP_302_FOUND
        assert "set-cookie" in r.headers.keys()
        cookie = find_cookie(searched_cookie_name="bearer", cookie_header=r.headers["set-cookie"])
        assert cookie is not None
        claim = decode_token(cookie.split("=")[1])
        assert claim["sub"] == uid

        # Check that user is created in RGW
        assert mock_rgw_admin.get_user(uid)["keys"][0]["user"] == uid

        # Check that user is created in DB
        db_user = (await db.execute(select(User).where(User.uid == uid))).scalar()
        assert db_user
        assert db_user.uid == uid

        # Cleanup
        await db.delete(db_user)
        await db.commit()
        mock_rgw_admin.delete_user(uid)


def find_cookie(searched_cookie_name: str, cookie_header: str) -> str | None:
    for cookie in cookie_header.split(";"):
        if cookie.startswith(searched_cookie_name):
            return cookie
    return None
