import pytest
from fastapi import status
from httpx import AsyncClient
from sqlalchemy.ext.asyncio import AsyncSession

from app.models.bucket import Bucket
from app.models.bucket_permission import BucketPermission, PermissionEnum
from app.models.user import User
from app.tests.mocks.mock_s3_resource import MockS3ServiceResource
from app.tests.utils.user import get_authorization_headers
from app.tests.utils.utils import random_lower_string


class _TestS3ObjectsRoutes:
    base_path = "/buckets/"


class TestS3ObjectsRoutesGet(_TestS3ObjectsRoutes):
    @pytest.mark.asyncio
    async def test_get_objects_with_right_for_specific_prefix(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_bucket: Bucket,
        random_second_user: User,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for getting the list of S3 objects in a bucket while only having rights for a specific prefix.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects. pytest fixture.
        """
        user_token_headers = get_authorization_headers(random_second_user.uid)
        mock_s3_service.create_object_in_bucket(bucket_name=random_bucket.name, key=random_lower_string())
        obj = mock_s3_service.create_object_in_bucket(
            bucket_name=random_bucket.name, key="pseudo/folder/" + random_lower_string()
        )
        permission = BucketPermission(
            bucket_name=random_bucket.name,
            user_id=random_second_user.uid,
            permissions=PermissionEnum.READ,
            file_prefix="pseudo/folder/",
        )
        db.add(permission)
        await db.commit()
        response = await client.get(f"{self.base_path}{random_bucket.name}/objects", headers=user_token_headers)
        assert response.status_code == status.HTTP_200_OK
        response_obj_list = response.json()
        assert len(response_obj_list) == 1
        assert response_obj_list[0]["key"] == obj.key

    @pytest.mark.asyncio
    async def test_get_object_without_right_for_specific_prefix(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_bucket: Bucket,
        random_second_user: User,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for getting a specific S3 object in a bucket while not having rights for the prefix.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects. pytest fixture.
        """
        user_token_headers = get_authorization_headers(random_second_user.uid)
        obj = mock_s3_service.create_object_in_bucket(
            bucket_name=random_bucket.name, key="another/folder/" + random_lower_string()
        )
        permission = BucketPermission(
            bucket_name=random_bucket.name,
            user_id=random_second_user.uid,
            permissions=PermissionEnum.READ,
            file_prefix="pseudo/folder/",
        )
        db.add(permission)
        await db.commit()
        response = await client.get(
            f"{self.base_path}{random_bucket.name}/objects/{obj.key}", headers=user_token_headers
        )
        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.asyncio
    async def test_get_all_s3_object_from_bucket(
        self,
        client: AsyncClient,
        random_bucket: Bucket,
        user_token_headers: dict[str, str],
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for getting the list of S3 objects in a bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects. pytest fixture.
        """
        # Create MockS3ObjectSummary in mock service
        obj = mock_s3_service.create_object_in_bucket(bucket_name=random_bucket.name, key=random_lower_string())

        response = await client.get(f"{self.base_path}{random_bucket.name}/objects", headers=user_token_headers)
        response_obj_list = response.json()

        assert response.status_code == status.HTTP_200_OK
        assert isinstance(response_obj_list, list)
        assert len(response_obj_list) > 0
        assert len(response_obj_list) == len(mock_s3_service.Bucket(name=random_bucket.name).get_objects())

        response_obj = response_obj_list[0]
        assert response_obj
        assert response_obj["key"] == obj.key
        assert response_obj["bucket"] == obj.bucket_name

    @pytest.mark.asyncio
    async def test_get_unknown_s3_object(
        self, client: AsyncClient, random_bucket: Bucket, user_token_headers: dict[str, str]
    ) -> None:
        """
        Test for getting an unknown object from a bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        """
        response = await client.get(
            f"{self.base_path}{random_bucket.name}/objects/impossible.pdf", headers=user_token_headers
        )

        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_get_s3_object(
        self,
        client: AsyncClient,
        random_bucket: Bucket,
        user_token_headers: dict[str, str],
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for getting a specific object from a bucket.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects. pytest fixture.
        """
        # Create MockS3ObjectSummary in mock service
        obj = mock_s3_service.create_object_in_bucket(bucket_name=random_bucket.name, key=random_lower_string())
        response = await client.get(
            f"{self.base_path}{random_bucket.name}/objects/{obj.key}", headers=user_token_headers
        )
        response_obj = response.json()
        assert response.status_code == status.HTTP_200_OK
        assert response_obj
        assert response_obj["key"] == obj.key
        assert response_obj["bucket"] == obj.bucket_name
