import random

import pytest
from fastapi import status
from httpx import AsyncClient

from app.models.user import User
from app.tests.utils.user import get_authorization_headers


class _TestUserRoutes:
    base_path = "/users/"


class TestUserRoutesGet(_TestUserRoutes):
    @pytest.mark.asyncio
    async def test_get_user_me(self, client: AsyncClient, random_user: User) -> None:
        """
        Test for getting the currently logged-in user.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_user : app.models.user.User
            Random user for testing. pytest fixture.
        """
        headers = get_authorization_headers(random_user.uid)
        response = await client.get(f"{self.base_path}me", headers=headers)
        current_user = response.json()
        assert response.status_code == status.HTTP_200_OK
        assert current_user
        assert current_user["uid"] == random_user.uid
        assert current_user["display_name"] == random_user.display_name

    @pytest.mark.asyncio
    async def test_get_unknown_user(self, client: AsyncClient, user_token_headers: dict[str, str]) -> None:
        """
        Test for getting an unknown user by its uid.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        """
        response = await client.get(f"{self.base_path}impossible_uid", headers=user_token_headers)
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_get_user_by_uid(self, client: AsyncClient, random_user: User) -> None:
        """
        Test for getting a known user by its uid.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_user : app.models.user.User
            Random user for testing. pytest fixture.
        """
        headers = get_authorization_headers(random_user.uid)
        response = await client.get(f"{self.base_path}{random_user.uid}", headers=headers)
        current_user = response.json()
        assert response.status_code == status.HTTP_200_OK
        assert current_user
        assert current_user["uid"] == random_user.uid
        assert current_user["display_name"] == random_user.display_name

    @pytest.mark.asyncio
    async def test_get_foreign_user_by_uid(
        self,
        client: AsyncClient,
        user_token_headers: dict[str, str],
        random_second_user: User,
    ) -> None:
        """
        Test for getting a foreign user by its uid.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        user_token_headers : dict[str,str]
            HTTP Headers to authorize the request. pytest fixture.
        random_second_user : app.models.user.User
            Random foreign user for testing. pytest fixture.
        """
        response = await client.get(f"{self.base_path}{random_second_user.uid}", headers=user_token_headers)

        assert response.status_code == status.HTTP_403_FORBIDDEN

    @pytest.mark.asyncio
    async def test_search_user_by_name_substring(
        self, client: AsyncClient, random_user: User, user_token_headers: dict[str, str]
    ) -> None:
        """
        Test for searching a user by its name

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on. pytest fixture.
        random_user : app.models.user.User
            Random user for testing. pytest fixture.
        """
        substring_indices = [0, 0]
        while substring_indices[1] - substring_indices[0] < 3:
            substring_indices = sorted(random.choices(range(len(random_user.display_name)), k=2))

        random_substring = random_user.display_name[substring_indices[0] : substring_indices[1]]

        response = await client.get(
            f"{self.base_path}", params={"name_like": random_substring}, headers=user_token_headers
        )
        users = response.json()
        assert response.status_code == status.HTTP_200_OK
        assert len(users) > 0
        assert sum(1 for u in users if u["uid"] == random_user.uid) == 1
