import asyncio
import json
from typing import AsyncGenerator, Generator

import pytest
import pytest_asyncio
from httpx import AsyncClient
from sqlalchemy.ext.asyncio import AsyncSession

from app.api.dependencies import LoginException, get_rgw_admin, get_s3_resource, get_userinfo_from_access_token
from app.db.session import SessionAsync as Session
from app.main import app
from app.models.bucket import Bucket
from app.models.bucket_permission import BucketPermission as BucketPermissionDB
from app.models.user import User
from app.schemas.bucket_permission import BucketPermissionOut as BucketPermissionSchema
from app.tests.mocks.mock_rgw_admin import MockRGWAdmin
from app.tests.mocks.mock_s3_resource import MockS3ServiceResource
from app.tests.utils.bucket import create_random_bucket
from app.tests.utils.user import create_random_user, get_authorization_headers


@pytest.fixture(scope="session")
def event_loop() -> Generator:
    """
    Creates an instance of the default event loop for the test session.
    """
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def mock_rgw_admin() -> MockRGWAdmin:
    """
    Fixture for creating a mock object for the rgwadmin package.
    """
    return MockRGWAdmin()


@pytest.fixture(scope="session")
def mock_s3_service() -> MockS3ServiceResource:
    """
    Fixture for creating a mock object for the rgwadmin package.
    """
    return MockS3ServiceResource()


@pytest_asyncio.fixture(scope="module")
async def client(mock_rgw_admin: MockRGWAdmin, mock_s3_service: MockS3ServiceResource) -> AsyncGenerator:
    """
    Fixture for creating a TestClient and perform HTTP Request on it.
    Overrides the dependency for the RGW admin operations.
    """

    def get_mock_rgw() -> MockRGWAdmin:
        return mock_rgw_admin

    def get_mock_s3() -> MockS3ServiceResource:
        return mock_s3_service

    def get_mock_userinfo(voperson_id: str, name: str, error: bool = False) -> dict[str, str]:
        if error:
            raise LoginException(error_source="mock_error")
        return {"voperson_id": voperson_id + "@lifescience-ri.eu", "name": name}

    app.dependency_overrides[get_rgw_admin] = get_mock_rgw
    app.dependency_overrides[get_s3_resource] = get_mock_s3
    app.dependency_overrides[get_userinfo_from_access_token] = get_mock_userinfo
    async with AsyncClient(app=app, base_url="http://localhost:8000") as ac:
        yield ac
    app.dependency_overrides = {}


@pytest_asyncio.fixture(scope="module")
async def user_token_headers(random_user: User) -> dict[str, str]:
    """
    Create valid authorization header with a successful login.
    """
    return get_authorization_headers(uid=random_user.uid)


@pytest_asyncio.fixture(scope="module")
async def db() -> AsyncGenerator:
    """
    Fixture for creating a database session to connect to.
    """
    async with Session() as dbSession:
        yield dbSession


@pytest_asyncio.fixture(scope="module")
async def random_user(db: AsyncSession, mock_rgw_admin: MockRGWAdmin) -> AsyncGenerator:
    """
    Create a random user and deletes him afterwards.
    """
    user = await create_random_user(db)
    mock_rgw_admin.create_key(uid=user.uid)
    yield user
    mock_rgw_admin.delete_user(uid=user.uid)
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="module")
async def random_second_user(db: AsyncSession, mock_rgw_admin: MockRGWAdmin) -> AsyncGenerator:
    """
    Create a random second user and deletes him afterwards.
    """
    user = await create_random_user(db)
    mock_rgw_admin.create_key(uid=user.uid)
    yield user
    mock_rgw_admin.delete_user(uid=user.uid)
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="module")
async def random_third_user(db: AsyncSession, mock_rgw_admin: MockRGWAdmin) -> AsyncGenerator:
    """
    Create a random third user and deletes him afterwards.
    """
    user = await create_random_user(db)
    mock_rgw_admin.create_key(uid=user.uid)
    yield user
    mock_rgw_admin.delete_user(uid=user.uid)
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_bucket(db: AsyncSession, random_user: User, mock_s3_service: MockS3ServiceResource) -> AsyncGenerator:
    """
    Create a random user and deletes him afterwards.
    """
    bucket = await create_random_bucket(db, random_user)
    mock_s3_service.Bucket(name=bucket.name).create()
    mock_s3_service.BucketPolicy(bucket.name).put(
        json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "PseudoOwnerPerm",
                        "Effect": "Allow",
                        "Principal": {"AWS": [f"arn:aws:iam:::user/{random_user.uid}"]},
                        "Action": ["s3:GetObject", "s3:DeleteObject", "s3:PutObject", "s3:ListBucket"],
                        "Resource": [f"arn:aws:s3:::{bucket.name}/*", f"arn:aws:s3:::{bucket.name}"],
                    }
                ],
            }
        )
    )
    yield bucket
    mock_s3_service.delete_bucket(name=bucket.name, force_delete=True)
    await db.delete(bucket)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_bucket_permission(
    db: AsyncSession, random_second_user: User, random_bucket: Bucket, mock_s3_service: MockS3ServiceResource
) -> BucketPermissionDB:
    """
    Create a bucket READ permission for the second user on a bucket.
    """
    permission_db = BucketPermissionDB(user_id=random_second_user.uid, bucket_name=random_bucket.name)
    db.add(permission_db)
    await db.commit()
    await db.refresh(permission_db)
    mock_s3_service.Bucket(random_bucket.name).Policy().put(
        json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": BucketPermissionSchema.from_db_model(permission_db).map_to_bucket_policy_statement(
                    random_second_user.uid
                ),
            }
        )
    )
    return permission_db


@pytest_asyncio.fixture(scope="function")
async def random_bucket_permission_schema(
    random_bucket_permission: BucketPermissionDB, random_second_user: User
) -> BucketPermissionSchema:
    """
    Create a bucket READ permission for the second user on a bucket.
    """

    return BucketPermissionSchema.from_db_model(random_bucket_permission, random_second_user.uid)


@pytest_asyncio.fixture(autouse=True)
async def multiple_random_users_buckets(
    db: AsyncSession, mock_rgw_admin: MockRGWAdmin, mock_s3_service: MockS3ServiceResource
) -> AsyncGenerator:
    """
    Create multiple random users and buckets to ensure that the database is not empty.
    """
    user1 = await create_random_user(db)
    mock_rgw_admin.create_key(uid=user1.uid)
    user2 = await create_random_user(db)
    mock_rgw_admin.create_key(uid=user2.uid)
    bucket1 = await create_random_bucket(db, user1)
    bucket2 = await create_random_bucket(db, user2)
    mock_s3_service.Bucket(name=bucket1.name).create()
    mock_s3_service.Bucket(name=bucket2.name).create()
    yield
    mock_rgw_admin.delete_user(uid=user1.uid)
    mock_rgw_admin.delete_user(uid=user2.uid)
    await db.delete(user1)
    await db.delete(user2)
    await db.delete(bucket1)
    await db.delete(bucket2)
    mock_s3_service.delete_bucket(name=bucket1.name)
    mock_s3_service.delete_bucket(name=bucket2.name)
