from datetime import datetime, timedelta

import pytest
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from app.crud.crud_bucket import CRUDBucket
from app.models.bucket import Bucket
from app.models.bucket_permission import PermissionEnum
from app.models.user import User
from app.schemas.bucket import BucketIn
from app.tests.utils.bucket import add_permission_for_bucket
from app.tests.utils.utils import random_lower_string


class TestBucketCRUDGet:
    @pytest.mark.asyncio
    async def test_get_bucket_by_name(self, db: AsyncSession, random_bucket: Bucket) -> None:
        """
        Test for getting a user by id from the User CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        bucket = await CRUDBucket.get(db, random_bucket.name)
        assert bucket
        assert bucket.name == random_bucket.name
        assert bucket.public == random_bucket.public
        assert bucket.description == random_bucket.description

    @pytest.mark.asyncio
    async def test_get_unknown_bucket(self, db: AsyncSession) -> None:
        """
        Test for getting a not existing bucket from the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        """
        bucket = await CRUDBucket.get(db, "unknown Bucket")
        assert bucket is None

    @pytest.mark.asyncio
    async def test_get_own_buckets(self, db: AsyncSession, random_bucket: Bucket) -> None:
        """
        Test for getting only the buckets where a user is the owner from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        buckets = await CRUDBucket.get_own_buckets(db, random_bucket.owner_id)

        assert len(buckets) == 1
        assert buckets[0].name == random_bucket.name

    @pytest.mark.asyncio
    async def test_get_bucket_with_read_permission_and_own(
        self,
        db: AsyncSession,
        random_bucket: Bucket,
        random_user: User,
        random_second_user: User,
    ) -> None:
        """
        Test for getting the users own bucket and a foreign bucket with READ permissions from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_user : app.models.user.User
            Random user for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        bucket = Bucket(
            name=random_lower_string(),
            description=random_lower_string(127),
            owner_id=random_second_user.uid,
        )
        db.add(bucket)
        await db.commit()
        await add_permission_for_bucket(db, bucket.name, random_user.uid, permission=PermissionEnum.READ)

        buckets = await CRUDBucket.get_for_user(db, random_user.uid)

        assert len(buckets) == 2
        assert buckets[0].name == random_bucket.name or buckets[1].name == random_bucket.name
        assert buckets[0].name == bucket.name or buckets[1].name == bucket.name

        await db.delete(bucket)

    @pytest.mark.asyncio
    async def test_get_bucket_with_read_permission(
        self, db: AsyncSession, random_bucket: Bucket, random_second_user: User
    ) -> None:
        """
        Test for getting a foreign bucket with READ permissions from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        await add_permission_for_bucket(db, random_bucket.name, random_second_user.uid, permission=PermissionEnum.READ)

        buckets = await CRUDBucket.get_for_user(db, random_second_user.uid)

        assert len(buckets) > 0
        assert buckets[0].name == random_bucket.name

    @pytest.mark.asyncio
    async def test_get_bucket_with_readwrite_permission(
        self, db: AsyncSession, random_bucket: Bucket, random_second_user: User
    ) -> None:
        """
        Test for getting a foreign bucket with READWRITE permissions from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        await add_permission_for_bucket(
            db, random_bucket.name, random_second_user.uid, permission=PermissionEnum.READWRITE
        )

        buckets = await CRUDBucket.get_for_user(db, random_second_user.uid)

        assert len(buckets) > 0
        assert buckets[0].name == random_bucket.name

    @pytest.mark.asyncio
    async def test_get_bucket_with_write_permission(
        self, db: AsyncSession, random_bucket: Bucket, random_second_user: User
    ) -> None:
        """
        Test for getting a foreign bucket with WRITE permissions from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        await add_permission_for_bucket(db, random_bucket.name, random_second_user.uid, permission=PermissionEnum.WRITE)

        buckets = await CRUDBucket.get_for_user(db, random_second_user.uid)

        assert len(buckets) == 0

    @pytest.mark.asyncio
    async def test_get_bucket_with_valid_time_permission(
        self, db: AsyncSession, random_bucket: Bucket, random_second_user: User
    ) -> None:
        """
        Test for getting a foreign bucket with valid time permission from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        await add_permission_for_bucket(
            db,
            random_bucket.name,
            random_second_user.uid,
            from_=datetime.now() - timedelta(days=10),
            to=datetime.now() + timedelta(days=10),
        )

        buckets = await CRUDBucket.get_for_user(db, random_second_user.uid)

        assert len(buckets) > 0
        assert buckets[0].name == random_bucket.name

    @pytest.mark.asyncio
    async def test_get_bucket_with_invalid_from_time_permission(
        self, db: AsyncSession, random_bucket: Bucket, random_second_user: User
    ) -> None:
        """
        Test for getting a foreign bucket with invalid 'from' time permission from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        await add_permission_for_bucket(
            db, random_bucket.name, random_second_user.uid, from_=datetime.now() + timedelta(days=10)
        )

        buckets = await CRUDBucket.get_for_user(db, random_second_user.uid)

        assert len(buckets) == 0

    @pytest.mark.asyncio
    async def test_get_bucket_with_invalid_to_time_permission(
        self, db: AsyncSession, random_bucket: Bucket, random_second_user: User
    ) -> None:
        """
        Test for getting a foreign bucket with invalid 'to' time permission from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        """
        await add_permission_for_bucket(
            db, random_bucket.name, random_second_user.uid, to=datetime.now() - timedelta(days=10)
        )

        buckets = await CRUDBucket.get_for_user(db, random_second_user.uid)

        assert len(buckets) == 0


class TestBucketCRUDCreate:
    @pytest.mark.asyncio
    async def test_create_bucket(self, db: AsyncSession, random_user: User) -> None:
        """
        Test for creating a bucket with the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_user : app.models.user.User
            Random user for testing. pytest fixture.
        """
        bucket_info = BucketIn(name=random_lower_string(), description=random_lower_string(127))
        bucket = await CRUDBucket.create(db, bucket_info, random_user.uid)
        assert bucket
        assert bucket.name == bucket_info.name
        assert bucket.owner_id == random_user.uid
        assert bucket.description == bucket_info.description

        stmt = select(Bucket).where(Bucket.name == bucket_info.name)
        bucket_db = (await db.execute(stmt)).scalar()

        assert bucket_db
        assert bucket_db.name == bucket_info.name
        assert bucket_db.owner_id == random_user.uid
        assert bucket_db.description == bucket_info.description

        await db.delete(bucket)

    @pytest.mark.asyncio
    async def test_create_duplicated_bucket(self, db: AsyncSession, random_bucket: Bucket) -> None:
        """
        Test for creating a duplicated bucket with the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        bucket_info = BucketIn(name=random_bucket.name, description=random_lower_string(127))
        bucket = await CRUDBucket.create(db, bucket_info, random_bucket.owner_id)

        assert bucket is None


class TestBucketCRUDDelete:
    @pytest.mark.asyncio
    async def test_delete_bucket(self, db: AsyncSession, random_bucket: Bucket) -> None:
        """
        Test for deleting a bucket with the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        await CRUDBucket.delete(db, random_bucket)

        stmt = select(Bucket).where(Bucket.name == random_bucket.name)
        bucket_db = (await db.execute(stmt)).scalar()

        assert bucket_db is None
