from datetime import datetime, timedelta

import pytest
from sqlalchemy import and_
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from app.crud.crud_bucket_permission import CRUDBucketPermission, DuplicateError
from app.models.bucket import Bucket
from app.models.bucket_permission import BucketPermission as BucketPermissionDB
from app.models.bucket_permission import PermissionEnum
from app.models.user import User
from app.schemas.bucket_permission import BucketPermissionIn as BucketPermissionSchema
from app.schemas.bucket_permission import BucketPermissionParameters as BucketPermissionParametersSchema


class TestBucketPermissionCRUDGet:
    @pytest.mark.asyncio
    async def test_get_specific_bucket_permission(
        self, db: AsyncSession, random_bucket_permission: BucketPermissionDB
    ) -> None:
        """
        Test for getting a specific bucket permission from the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket_permission : app.models.bucket_permission.BucketPermission
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        bucket_permission = await CRUDBucketPermission.get(
            db, bucket_name=random_bucket_permission.bucket_name, user_id=random_bucket_permission.user_id
        )
        assert bucket_permission
        assert bucket_permission.user_id == random_bucket_permission.user_id
        assert bucket_permission.bucket_name == random_bucket_permission.bucket_name
        assert bucket_permission.permissions == random_bucket_permission.permissions

    @pytest.mark.asyncio
    async def test_get_bucket_permissions_by_bucket_name(
        self, db: AsyncSession, random_bucket_permission: BucketPermissionDB
    ) -> None:
        """
        Test for getting all bucket permissions for a specific bucket from the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket_permission : app.models.bucket_permission.BucketPermission
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        bucket_permissions = await CRUDBucketPermission.get_permissions_for_bucket(
            db, bucket_name=random_bucket_permission.bucket_name
        )
        assert len(bucket_permissions) == 1
        bucket_permission = bucket_permissions[0]
        assert bucket_permission.user_id == random_bucket_permission.user_id
        assert bucket_permission.bucket_name == random_bucket_permission.bucket_name
        assert bucket_permission.permissions == random_bucket_permission.permissions

    @pytest.mark.asyncio
    async def test_get_bucket_permissions_by_uid(
        self, db: AsyncSession, random_bucket_permission: BucketPermissionDB
    ) -> None:
        """
        Test for getting all bucket permissions for a specific user from the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket_permission : app.models.bucket_permission.BucketPermission
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        bucket_permissions = await CRUDBucketPermission.get_permissions_for_user(
            db, user_id=random_bucket_permission.user_id
        )
        assert len(bucket_permissions) == 1
        bucket_permission = bucket_permissions[0]
        assert bucket_permission.user_id == random_bucket_permission.user_id
        assert bucket_permission.bucket_name == random_bucket_permission.bucket_name
        assert bucket_permission.permissions == random_bucket_permission.permissions


class TestBucketPermissionCRUDCreate:
    @pytest.mark.asyncio
    async def test_create_bucket_permissions_for_unknown_user(self, db: AsyncSession, random_bucket: Bucket) -> None:
        """
        Test for creating a bucket permission for an unknown user.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        permission = BucketPermissionSchema(bucket_name=random_bucket.name, uid="ImpossibleUser")
        with pytest.raises(KeyError):
            await CRUDBucketPermission.create(db, permission)

    @pytest.mark.asyncio
    async def test_create_bucket_permissions_for_owner(
        self, db: AsyncSession, random_user: User, random_bucket: Bucket
    ) -> None:
        """
        Test for creating a bucket permission for the owner of the bucket.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_user : app.models.user.User
            Random user for testing who is owner of the bucket. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        permission = BucketPermissionSchema(bucket_name=random_bucket.name, uid=random_user.uid)
        with pytest.raises(ValueError):
            await CRUDBucketPermission.create(db, permission)

    @pytest.mark.asyncio
    async def test_create_duplicate_bucket_permissions(
        self, db: AsyncSession, random_bucket_permission_schema: BucketPermissionSchema
    ) -> None:
        """
        Test for creating a duplicated bucket permission.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket_permission_schema : app.schemas.bucket_permission.BucketPermissionOut
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        permission = BucketPermissionSchema(
            bucket_name=random_bucket_permission_schema.bucket_name, uid=random_bucket_permission_schema.uid
        )
        with pytest.raises(DuplicateError):
            await CRUDBucketPermission.create(db, permission)

    @pytest.mark.asyncio
    async def test_create_valid_bucket_permissions(
        self, db: AsyncSession, random_second_user: User, random_bucket: Bucket
    ) -> None:
        """
        Test for creating a valid bucket permission.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_second_user : app.models.user.User
            Random second user for testing. pytest fixture.
        random_bucket : app.models.bucket.Bucket
            Random bucket for testing. pytest fixture.
        """
        permission = BucketPermissionSchema(bucket_name=random_bucket.name, uid=random_second_user.uid)
        created_permission = await CRUDBucketPermission.create(db, permission)

        assert created_permission.user_id == random_second_user.uid
        assert created_permission.bucket_name == random_bucket.name


class TestBucketPermissionCRUDDelete:
    @pytest.mark.asyncio
    async def test_delete_bucket_permissions(
        self, db: AsyncSession, random_bucket_permission: BucketPermissionDB
    ) -> None:
        """
        Test for creating a valid bucket permission.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket_permission : app.models.bucket_permission.BucketPermission
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        await CRUDBucketPermission.delete(db, random_bucket_permission)

        stmt = select(BucketPermissionDB).where(
            and_(
                BucketPermissionDB.bucket_name == random_bucket_permission.bucket_name,
                BucketPermissionDB.user_id == random_bucket_permission.user_id,
            )
        )
        bucket_permission_db = (await db.execute(stmt)).scalar()

        assert bucket_permission_db is None


class TestBucketPermissionCRUDUpdate:
    @pytest.mark.asyncio
    async def test_update_bucket_permissions(
        self, db: AsyncSession, random_bucket_permission: BucketPermissionDB
    ) -> None:
        """
        Test for updating a valid bucket permission in the database.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_bucket_permission : app.models.bucket_permission.BucketPermission
            Bucket permission for a random bucket for testing. pytest fixture.
        """
        new_from_time = datetime(2022, 1, 1, 0, 0)
        new_params = BucketPermissionParametersSchema(
            from_timestamp=new_from_time,
            to_timestamp=new_from_time + timedelta(days=1),
            permission=PermissionEnum.READWRITE,
            file_prefix="pseudo/folder/",
        )
        new_permission = await CRUDBucketPermission.update_permission(db, random_bucket_permission, new_params)

        assert new_permission.user_id == random_bucket_permission.user_id
        assert new_permission.bucket_name == random_bucket_permission.bucket_name
        assert new_permission.from_ == new_params.from_timestamp
        assert new_permission.to == new_params.to_timestamp
        assert new_permission.permissions == new_params.permission
        assert new_permission.file_prefix == new_params.file_prefix
