import random

import pytest
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from app.crud.crud_user import CRUDUser
from app.models.user import User
from app.tests.utils.utils import random_lower_string


class TestUserCRUD:
    @pytest.mark.asyncio
    async def test_create_user(self, db: AsyncSession) -> None:
        """
        Test for creating a user in the User CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        """
        user = User(uid=random_lower_string(), display_name=random_lower_string())
        await CRUDUser.create(db, user)

        db_user = (await db.execute(select(User).where(User.uid == user.uid))).scalar()
        assert db_user
        assert db_user.uid == user.uid

        await db.delete(db_user)
        await db.commit()

    @pytest.mark.asyncio
    async def test_get_user_by_id(self, db: AsyncSession, random_user: User) -> None:
        """
        Test for getting a user by id from the User CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_user : app.models.user.User
            Random user for testing. pytest fixture.
        """
        user = await CRUDUser.get(db, random_user.uid)
        assert user
        assert random_user.uid == user.uid
        assert random_user.display_name == user.display_name

    @pytest.mark.asyncio
    async def test_get_unknown_user_by_id(
        self,
        db: AsyncSession,
    ) -> None:
        """
        Test for getting an unknown user by id from the User CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        """
        user = await CRUDUser.get(db, random_lower_string(length=16))
        assert user is None

    @pytest.mark.asyncio
    async def test_search_successful_user_by_name(self, db: AsyncSession, random_user: User) -> None:
        """
        Test for searching a user by a substring of his name in the User CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_user : app.models.user.User
            Random user for testing. pytest fixture.
        """
        substring_indices = [0, 0]
        while substring_indices[0] == substring_indices[1]:
            substring_indices = sorted(random.choices(range(len(random_user.display_name)), k=2))

        random_substring = random_user.display_name[substring_indices[0] : substring_indices[1]]
        users = await CRUDUser.search_for_name(db, random_substring)
        assert len(users) > 0
        assert sum(1 for u in users if u.uid == random_user.uid) == 1

    @pytest.mark.asyncio
    async def test_search_non_existing_user_by_name(self, db: AsyncSession, random_user: User) -> None:
        """
        Test for searching a non-existing user by a substring of his name in the User CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on. pytest fixture.
        random_user : app.models.user.User
            Random user for testing. pytest fixture.
        """
        users = await CRUDUser.search_for_name(db, 2 * random_user.display_name)
        assert sum(1 for u in users if u.uid == random_user.uid) == 0
