from datetime import datetime

from botocore.exceptions import ClientError


class MockS3Object:
    """
    Mock S3 object for the boto3 S3Object for testing purposes.

    Attributes
    ----------
    key : str
        Key of the S3 object.
    bucket_name : str
        Name of the corresponding bucket.
    """

    def __init__(self, bucket_name: str, key: str) -> None:
        """
        Initialize a MockS3Object.

        Parameters
        ----------
        bucket_name : str
            Name of the corresponding bucket.
        key : str
            Key of the S3 object.
        """
        self.key = key
        self.bucket_name = bucket_name
        self.content_type = "text/plain"

    def __repr__(self) -> str:
        return f"MockS3Object(key={self.key}, bucket={self.bucket_name})"


class MockS3ObjectSummary:
    """
    Mock S3 object summary for the boto3 S3ObjectSummary for testing purposes.

    Functions
    ---------
    Object() -> MockS3Object
        Save a new bucket policy.

    Attributes
    ----------
    key : str
        Key of the S3 object.
    bucket_name : str
        Name of the corresponding bucket.
    size : int
        Size of object in bytes. Always 100.
    last_modified : datetime
        Time and date of last modification of this object.
    """

    def __init__(self, bucket_name: str, key: str) -> None:
        """
        Initialize a MockS3ObjectSummary.

        Parameters
        ----------
        bucket_name : str
            Name of the corresponding bucket.
        key : str
            Key of the S3 object.
        """
        self.key = key
        self.bucket_name = bucket_name
        self.size = 100
        self.last_modified = datetime.now()

    def __repr__(self) -> str:
        return f"MockS3ObjectSummary(key={self.key}, bucket={self.bucket_name})"

    def Object(self) -> MockS3Object:
        """
        Get the S3 Object from the summary.

        Returns
        -------
        sObject : app.tests.mocks.mock_s3_resource.MockS3Object
            The corresponding S3Object.
        """
        return MockS3Object(self.bucket_name, self.key)


class MockS3BucketPolicy:
    """
    Mock S3 bucket policy the boto3 BucketPolicy for testing purposes.

    Functions
    ---------
    put(Policy: str) -> None
        Save a new bucket policy.

    Attributes
    ----------
    bucket_name : str
        Name of the corresponding bucket.
    policy : str
        The policy in as a string.
    """

    def __init__(self, bucket_name: str):
        self.bucket_name = bucket_name
        self.policy: str = ""

    def put(self, Policy: str) -> None:
        """
        Save a new bucket policy.

        Parameters
        ----------
        Policy : str
            The new policy as str.
        """
        self.policy = Policy


class MockS3Bucket:
    """
    Mock S3 bucket for the boto3 Bucket for testing purposes.

    Functions
    ---------
    Policy() -> app.tests.mocks.mock_s3_resource.MockS3BucketPolicy
        Get the bucket policy from the bucket.
    create() -> None
        Create the bucket in the mock service.
    delete() -> None
        Delete the bucket in the mock service
    delete_objects(Delete: dict[str, list[dict[str, str]]]) -> None
        Delete multiple objects in the bucket.
    get_objects() -> list[app.tests.mocks.mock_s3_resource.MockS3ObjectSummary]
        List of MockS3ObjectSummary in the bucket.
    add_object(obj: app.tests.mocks.mock_s3_resource.MockS3ObjectSummary) -> None
        Add a MockS3ObjectSummary to the bucket.

    Attributes
    ----------
    name: str
        name of the bucket.
    creation_date : datetime
        Creation date of the bucket.
    objects : app.tests.mocks.mock_s3_resource.MockS3Bucket.MockS3ObjectList
        Object in the bucket in a proxy class.
    """

    class MockS3ObjectList:
        """
        Proxy object to package a list in a class.

        Important because you have to access all the S3ObjectsSummary of a bucket with
        S3Bucket.objects.all()

        Functions
        ---------
        all() -> list[app.tests.mocks.mock_s3_resource.MockS3ObjectSummary]
            Get the saved list.
        filter(Prefix: str) -> app.tests.mocks.mock_s3_resource.MockS3Bucket.MockS3ObjectList
            Filter the object in the list by the prefix all their keys should have.
        add(obj: app.tests.mocks.mock_s3_resource.MockS3ObjectSummary) -> None
            Add a MockS3ObjectSummary to the list.
        delete(key: str) -> None
            Delete a MockS3ObjectSummary from the list
        """

        def __init__(self, obj_list: list[MockS3ObjectSummary] | None = None) -> None:
            self._objs: list[MockS3ObjectSummary] = [] if obj_list is None else obj_list

        def all(self) -> list[MockS3ObjectSummary]:
            """
            Get the saved list.

            Returns
            -------
            objects : list[app.tests.mocks.mock_s3_resource.MockS3ObjectSummary]
                List of MockS3ObjectSummary
            """
            return self._objs

        def add(self, obj: MockS3ObjectSummary) -> None:
            """
            Add a MockS3ObjectSummary to the list.

            Parameters
            ----------
            obj : app.tests.mocks.mock_s3_resource.MockS3ObjectSummary
                MockS3ObjectSummary which should be added to the list
            """
            self._objs.append(obj)

        def delete(self, key: str) -> None:
            """
            Delete a MockS3ObjectSummary from the list.

            Parameters
            ----------
            key : str
                Key of the object to delete
            """
            self._objs = [obj for obj in self._objs if obj.key != key]

        def filter(self, Prefix: str) -> "MockS3Bucket.MockS3ObjectList":
            """
            Filter the object in the list by the prefix all their keys should have.

            Parameters
            ----------
            Prefix : str
                The prefix that all keys should have.

            Returns
            -------
            obj_list : app.tests.mocks.mock_s3_resource.MockS3Bucket.MockS3ObjectList
                The filtered list.
            """
            return MockS3Bucket.MockS3ObjectList(obj_list=list(filter(lambda x: x.key.startswith(Prefix), self._objs)))

    def __init__(self, name: str, parent_service: "MockS3ServiceResource"):
        """
        Initialize a MockS3Bucket.

        Parameters
        ----------
        name : str
            Name of the bucket.
        parent_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock service object where the bucket will be saved.
        """
        self.name = name
        self.creation_date: datetime = datetime.now()
        self.objects = MockS3Bucket.MockS3ObjectList()
        self._parent_service: MockS3ServiceResource = parent_service
        self.policy = MockS3BucketPolicy(name)

    def Policy(self) -> MockS3BucketPolicy:
        """
        Get the bucket policy from the bucket.

        Returns
        -------
        bucket_policy : app.tests.mocks.mock_s3_resource.MockS3BucketPolicy
            The corresponding bucket policy.
        """
        return self.policy

    def create(self) -> None:
        """
        Create the bucket in the mock S3 service.
        """
        self._parent_service.create_bucket(self)

    def delete(self) -> None:
        """
        Delete the bucket in the mock S3 service.
        """
        self._parent_service.delete_bucket(self.name)

    def delete_objects(self, Delete: dict[str, list[dict[str, str]]]) -> None:
        """
        Delete multiple objects in the bucket.

        Parameters
        ----------
        Delete : dict[str, list[dict[str, str]]]
            The keys of the objects to delete.

        Notes
        -----
        The `Delete` parameter has the form
        {
            'Objects': [
                { 'Key' : 'test.txt' },...
            ]
        }
        """
        for key_object in Delete["Objects"]:
            self.objects.delete(key=key_object["Key"])

    def get_objects(self) -> list[MockS3ObjectSummary]:
        """
        Get the MockS3ObjectSummary in the bucket.
        Convenience function for testing.

        Returns
        -------
        objects : list[app.tests.mocks.mock_s3_resource.MockS3ObjectSummary]
            List of MockS3ObjectSummary in the bucket.
        """
        return self.objects.all()

    def add_object(self, obj: MockS3ObjectSummary) -> None:
        """
        Add a MockS3ObjectSummary to the bucket.
        Convenience function for testing.

        Parameters
        ----------
        obj : app.tests.mocks.mock_s3_resource.MockS3ObjectSummary
            Object to add.
        """
        self.objects.add(obj)

    def __repr__(self) -> str:
        return f"MockS3Bucket(name={self.name}, objects={[obj.key for obj in self.get_objects()]})"


class MockS3ServiceResource:
    """
    A functional mock class of the boto3 S3ServiceResource for testing purposes.

    Functions
    ---------
    Bucket(name: str) -> app.tests.mocks.mock_s3_resource.MockS3bucket
        Get/Create a mock bucket.
    ObjectSummary(bucket_name: str, key: str) -> app.tests.mocks.mock_s3_resource.MockS3objectSummary
        Create a mock object summary.
    BucketPolicy(bucket_name: str) -> app.tests.mocks.mock_s3_resource.MockS3BucketPolicy
        Get the bucket policy from the corresponding bucket.
    BucketAcl(bucket_name: str) -> app.tests.mocks.mock_s3_resource.MockS3BucketAcl
        Get the bucket acl from the corresponding bucket.
    create_bucket(bucket: app.tests.mocks.mock_s3_resource.MockS3bucket) -> None
        Create a bucket in the mock service.
    delete_bucket(name: str) -> None
        Delete a bucket in the mock service.
    create_object_in_bucket(bucket_name: str, key: str) -> app.tests.mocks.mock_s3_resource.MockS3objectSummary
        Create an MockS3ObjectSummary in a bucket.
    """

    def __init__(self) -> None:
        self._buckets: dict[str, MockS3Bucket] = {}

    def Bucket(self, name: str) -> MockS3Bucket:
        """
        Get an existing bucket from the mock service or creat a new mock bucket but doesn't save it yet.
        Call `bucket.create()` on returned object for that.

        Parameters
        ----------
        name : str
            Name of the bucket.

        Returns
        -------
        bucket : app.tests.mocks.mock_s3_resource.MockS3bucket
            Existing/Created mock bucket.
        """
        return self._buckets[name] if name in self._buckets else MockS3Bucket(name=name, parent_service=self)

    def ObjectSummary(self, bucket_name: str, key: str) -> MockS3ObjectSummary:
        """
        Create a mock object summary.

        Parameters
        ----------
        bucket_name : str
            Name of the corresponding bucket.
        key : str
            Key of the S3 object.

        Returns
        -------
        obj : app.tests.mocks.mock_s3_resource.MockS3objectSummary
              Existing MockS3ObjectSummary.

        Notes
        -----
        Raises an `botocore.exceptions.ClientError` if the object doesn't exist in the bucket.
        For creating a MockS3ObjectSummary see `create_object_in_bucket`.
        """
        if key not in map(lambda obj: obj.key, self._buckets[bucket_name].get_objects()):
            raise ClientError(operation_name="Mock", error_response={})
        return MockS3ObjectSummary(bucket_name=bucket_name, key=key)

    def BucketPolicy(self, bucket_name: str) -> MockS3BucketPolicy:
        """
        Get the bucket policy from the corresponding bucket.

        Parameters
        ----------
        bucket_name : str
            Name of the bucket.

        Returns
        -------
        bucket_policy : app.tests.mocks.mock_s3_resource.MockS3BucketPolicy
            The corresponding bucket policy.
        """
        return self._buckets[bucket_name].policy

    def create_bucket(self, bucket: MockS3Bucket) -> None:
        """
        Create a bucket in the mock service.
        Convenience function for testing.

        Parameters
        ----------
        bucket : app.tests.mocks.mock_s3_resource.MockS3bucket
            Bucket which should be created.
        """
        self._buckets[bucket.name] = bucket

    def delete_bucket(self, name: str, force_delete: bool = False) -> None:
        """
        Delete am empty bucket in the mock service.
        Convenience function for testing.

        Parameters
        ----------
        name : str
            Name of the bucket.
        force_delete : bool, default False
            Force deletes even a non-empty bucket.
        """
        if name in self._buckets:
            if not force_delete and len(self._buckets[name].get_objects()) > 0:
                raise ClientError(operation_name="Mock", error_response={})
            del self._buckets[name]

    def create_object_in_bucket(self, bucket_name: str, key: str) -> MockS3ObjectSummary:
        """
        Create an MockS3ObjectSummary in a bucket.
        Convenience function for testing.

        Parameters
        ----------
        bucket_name : str
            Name of the bucket.
        key : str
            key of the S3 object.

        Returns
        -------
        obj : app.tests.mocks.mock_s3_resource.MockS3ObjectSummary
            Newly created MockS3ObjectSummary.
        """
        obj = MockS3ObjectSummary(bucket_name=bucket_name, key=key)
        self._buckets[bucket_name].add_object(obj)
        return obj

    def __repr__(self) -> str:
        return f"MockS3ServiceResource(buckets={[bucket_name for bucket_name in self._buckets.keys()]})"
