from datetime import datetime

import pytest
from sqlalchemy.ext.asyncio import AsyncSession

from app.models.bucket import Bucket
from app.models.bucket_permission import BucketPermission, PermissionEnum
from app.models.user import User

from .utils import random_lower_string


@pytest.mark.asyncio
async def create_random_bucket(db: AsyncSession, user: User) -> Bucket:
    """
    Creates a random bucket in the database.

    Parameters
    ----------
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on.
    user : app.models.user.User
        Owner of the bucket.

    Returns
    -------
    bucket : app.models.bucket.Bucket
        Newly created bucket.
    """
    bucket = Bucket(
        name=random_lower_string(),
        description=random_lower_string(length=127),
        owner_id=user.uid,
    )
    db.add(bucket)
    await db.commit()
    return bucket


@pytest.mark.asyncio
async def add_permission_for_bucket(
    db: AsyncSession,
    bucket_name: str,
    uid: str,
    from_: datetime | None = None,
    to: datetime | None = None,
    permission: PermissionEnum = PermissionEnum.READ,
) -> None:
    """
    Creates Permission to a bucket for a user in the database.

    Parameters
    ----------
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on.
    bucket_name : str
        name of the bucket.
    uid : str
        UID of the user who we grant the permission.
    from_ : datetime.datetime | None, default None
        Time when from when the permission should be active.
    to : datetime.datetime | None, default None
        Time till when the permissions should be active.
    permission : app.models.bucket_permission.PermissionEnum, default PermissionEnum.READ
        The permission the user is granted.
    """
    perm = BucketPermission(
        user_id=uid,
        bucket_name=bucket_name,
        from_=from_,
        to=to,
        permissions=permission.name,
    )
    db.add(perm)
    await db.commit()
