import pytest
from sqlalchemy.ext.asyncio import AsyncSession

from app.core.security import create_access_token
from app.models.user import User

from .utils import random_lower_string


@pytest.mark.asyncio
def get_authorization_headers(uid: str) -> dict[str, str]:
    """
    Login a user and return the correct headers for subsequent requests.

    Parameters
    ----------
    uid : str
        UID of the user who should be logged in.

    Returns
    -------
    headers : dict[str,str]
        HTTP Headers to authorize each request.
    """
    a_token = create_access_token(uid)
    headers = {"Authorization": f"Bearer {a_token}"}
    return headers


@pytest.mark.asyncio
async def create_random_user(db: AsyncSession) -> User:
    """
    Creates a random user in the database.

    Parameters
    ----------
    db : sqlalchemy.ext.asyncio.AsyncSession.
        Async database session to perform query on.

    Returns
    -------
    user : app.models.user.User
        Newly created user.
    """
    user = User(
        uid=random_lower_string(),
        display_name=random_lower_string(),
    )
    db.add(user)
    await db.commit()
    return user
